"""Handler for clearing the legend quest"""
from typing import Any
from . import story_level_id_selector
from ... import helper


def edit_legend_quest(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for clearing the legend quest"""
    stage_data = save_stats["legend_quest"]
    lengths = stage_data["Lengths"]
    total = lengths["stages"]
    progress = story_level_id_selector.select_level_progress(None, total=total, examples=["레벨 1", "레벨 2"])

    if progress == 0:
        stage_data["Value"]["clear_progress"][0][0] = 0
        stage_data["Value"]["clear_amount"][0][0] = [0] * len(stage_data["Value"]["clear_amount"][0][0])
    else:
        stage_id = progress - 1
        stage_data["Value"]["clear_progress"][0][0] = min(progress, total)
        stage_data["Value"]["clear_amount"][0][0][stage_id] = 1
        stage_data["Value"]["tries"][0][0][stage_id] = 1
        for i in range(stage_id):
            stage_data["Value"]["clear_amount"][0][0][i] = 1
            stage_data["Value"]["tries"][0][0][i] = 1
        for i in range(stage_id + 1, total):
            stage_data["Value"]["clear_amount"][0][0][i] = 0
            stage_data["Value"]["tries"][0][0][i] = 0
    
    if stage_data["Value"]["clear_progress"][0][0] == total:
        stage_data["Value"]["unlock_next"][0][1] = lengths["stars"] - 1

    save_stats["legend_quest"] = stage_data
    helper.colored_text("레전드 퀘스트 스테이지 설정 성공")
    return save_stats


"""Handler for unlocking the aku realm"""
from typing import Any
from ... import helper
from . import event_stages


def unlock_aku_realm(save_stats: dict[str, Any]) -> dict[str, Any]:
    """
    Unlock the aku realm

    Args:
        save_stats (dict[str, Any]): The save stats to edit

    Returns:
        dict[str, Any]: The edited save stats
    """
    stage_ids = [255, 256, 257, 258, 265, 266, 268]
    offset = 400
    for stage_id in stage_ids:
        save_stats["event_stages"] = event_stages.set_stage_data(
            save_stats["event_stages"],
            stage_id + offset,
            1,
            save_stats["event_stages"]["Lengths"],
            True,
        )
    helper.colored_text("&마계 영역이 성공적으로 잠금 해제되었습니다.")
    return save_stats



"""Module for editing catseyes"""
from typing import Any

from ... import csv_handler, game_data_getter, helper, item


def get_catseye_ids(is_jp: bool) -> list[int]:
    """Get the catseye ids"""

    file_data = game_data_getter.get_file_latest("DataLocal", "Gatyaitembuy.csv", is_jp)
    if file_data is None:
        helper.error_text("캣츠아이 ID를 가져오지 못했습니다.s")
        return []
    items = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
            ",",
        )[1:]
    )
    catseye_ids: dict[int, int] = {}
    for item_id, item_data in enumerate(items):
        category = item_data[6]
        if category == 5:
            index = item_data[7]
            catseye_ids[index] = item_id
    ids = sorted(catseye_ids.items(), key=lambda x: x[0])
    return [id[1] for id in ids]


def get_catseye_names(is_jp: bool) -> list[str]:
    """Get the catseye names"""

    file_data = game_data_getter.get_file_latest("resLocal", "GatyaitemName.csv", is_jp)
    if file_data is None:
        helper.error_text("캣츠아이 이름을 가져오지 못했습니다.")
        return []
    item_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        helper.get_text_splitter(is_jp),
    )
    catseye_names: list[str] = []
    for catseye_id in get_catseye_ids(is_jp):
        try:
            catseye_names.append(item_names[catseye_id][0])
        except IndexError:
            helper.error_text(f"에 대한 캣츠아이 이름을 가져오지 못했습니다. {catseye_id}")
    return catseye_names


def edit_catseyes(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing catseyes"""

    catseyes = item.IntItemGroup.from_lists(
        names=get_catseye_names(helper.check_data_is_jp(save_stats)),
        values=save_stats["catseyes"],
        maxes=9999,
        group_name="캣츠아이",
    )
    catseyes.edit()
    save_stats["catseyes"] = catseyes.get_values()
    return save_stats



"""Handler for editing main story treasures"""
from typing import Any, Optional

from ... import helper, user_input_handler, item, csv_handler, game_data_getter
from . import story_level_id_selector, main_story


def get_stages(is_jp: bool) -> Optional[list[list[list[int]]]]:
    """Get what stages belong to which treasure group"""

    treasures_values: list[list[list[int]]] = []
    file_data = game_data_getter.get_file_latest(
        "DataLocal", "treasureData0.csv", is_jp
    )
    if file_data is None:
        helper.error_text("가져오지 못했습니다. treasureData0.csv")
        return None
    eoc_treasures = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
        )
    )[11:22]
    file_data = game_data_getter.get_file_latest(
        "DataLocal", "treasureData1.csv", is_jp
    )
    if file_data is None:
        helper.error_text("가져오지 못했습니다. treasureData1.csv")
        return None
    itf_treasures = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
        )
    )[11:22]
    file_data = game_data_getter.get_file_latest(
        "DataLocal", "treasureData2_0.csv", is_jp
    )
    if file_data is None:
        helper.error_text("가져오지 못했습니다. treasureData2_0.csv")
        return None
    cotc_treasures = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
        )
    )[11:22]

    treasures_values.append(remove_negative_1(eoc_treasures))
    treasures_values.append(remove_negative_1(itf_treasures))
    treasures_values.append(remove_negative_1(cotc_treasures))
    return treasures_values


def remove_negative_1(data: list[list[int]]) -> list[list[int]]:
    """Remove items from a list that have a negative value of 1"""

    new_data = data.copy()
    for i, val in enumerate(data):
        if -1 in val:
            new_data[i] = new_data[i][:-1]
    return new_data


def get_names(is_jp: bool) -> Optional[list[list[list[str]]]]:
    """Get the names of all of the treasure groups"""

    names: list[list[list[str]]] = []
    if is_jp:
        country_code = "ja"
    else:
        country_code = "en"

    file_data = game_data_getter.get_file_latest(
        "resLocal", f"Treasure3_0_{country_code}.csv", is_jp
    )
    if file_data is None:
        helper.error_text(f"가져오지 못했습니다. Treasure3_0_{country_code}.csv")
        return None
    eoc_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )[:11]

    file_data = game_data_getter.get_file_latest(
        "resLocal", f"Treasure3_1_AfterFirstEncounter_{country_code}.csv", is_jp
    )
    if file_data is None:
        helper.error_text(
            f"가져오지 못했습니다. Treasure3_1_AfterFirstEncounter_{country_code}.csv"
        )
        return None
    itf_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )[:11]

    file_data = game_data_getter.get_file_latest(
        "resLocal", f"Treasure3_2_0_{country_code}.csv", is_jp
    )
    if file_data is None:
        helper.error_text(f"가져오지 못했습니다. Treasure3_2_0_{country_code}.csv")
        return None
    cotc_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )[:11]

    names.append(helper.copy_first_n(eoc_names, 0))
    names.append(helper.copy_first_n(itf_names, 0))
    names.append(helper.copy_first_n(cotc_names, 0))

    return names


def get_treasure_groups(is_jp: bool) -> Optional[dict[str, Any]]:
    """Get the names and stages of all of the treasure groups"""

    treasure_stages = get_stages(is_jp)
    treasure_names = get_names(is_jp)
    if treasure_stages is None or treasure_names is None:
        return None
    return {"names": treasure_names, "stages": treasure_stages}


def set_treasures(
    treasure_stats: list[list[int]], user_levels: list[int]
) -> list[list[int]]:
    """Set the treasure stats of a set of levels"""

    for i, level in enumerate(user_levels):
        if level == -1:
            continue
        if i > 2:
            i += 1
        treasure_stats[i] = [level] * 48 + [0]
    return treasure_stats


def set_specific_treasures(
    treasure_stats: list[list[int]], treasure_data: list[int], chapter_id: int
) -> list[list[int]]:
    """Set the treasure stats of specific treasures"""

    for i, stage in enumerate(treasure_data):
        if stage == -1:
            continue
        if i > 45:
            stage_id = i
        else:
            stage_id = 45 - i
        treasure_stats[chapter_id][stage_id] = stage
    return treasure_stats


def set_treasure_groups(
    treasures_stats: list[list[int]],
    treasure_grps: dict[str, Any],
    treasure_levels: list[Optional[int]],
    type_id: int,
    chapter_id: int,
) -> list[list[int]]:
    """Set the treasure stats of a group of treasures"""

    for i, treasure_level in enumerate(treasure_levels):
        stages = treasure_grps["stages"][type_id][i]
        for stage in stages:
            if treasure_level is not None:
                treasures_stats[chapter_id][stage] = treasure_level
    return treasures_stats


def treasure_groups(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing treasure groups"""

    treasure_grps = get_treasure_groups(helper.check_data_is_jp(save_stats))
    if treasure_grps is None:
        return save_stats
    treasures_stats = save_stats["treasures"]

    ids = user_input_handler.select_not_inc(main_story.CHAPTERS, "설정")

    for chapter_id in ids:
        helper.colored_text(f"챕터: &{main_story.CHAPTERS[chapter_id]}&")
        type_id = chapter_id // 3
        if chapter_id > 2:
            chapter_id += 1
        names = treasure_grps["names"][type_id]
        treasure_levels = [-1] * len(names)
        helper.colored_text("&0& = 없음, &1& = 낮음, &2& = 보통, &3& = 높음")
        treasure_levels = item.IntItemGroup.from_lists(
            names=names,
            values=None,
            maxes=None,
            group_name="보물",
        )
        treasure_levels.edit()

        treasures_stats = set_treasure_groups(
            treasures_stats,
            treasure_grps,
            treasure_levels.get_values_none(),
            type_id,
            chapter_id,
        )
    save_stats["treasures"] = treasures_stats

    return save_stats


def specific_stages(save_stats: dict[str, Any]):
    """Handler for editing specific stages"""

    treasure_stats = save_stats["treasures"]
    chapter_ids = story_level_id_selector.select_specific_chapters()

    choice = story_level_id_selector.get_option()
    for chapter_id in chapter_ids:
        chapter_id = main_story.format_story_id(chapter_id)
        stage_ids = story_level_id_selector.select_levels(chapter_id, choice)
        treasure_data = [-1] * 48
        treasure_data = user_input_handler.handle_all_at_once(
            stage_ids,
            False,
            treasure_data,
            list(range(1, 49)),
            "treasure level",
            "stage",
            "(&0&=없음, &1&=열등, &2&=정상, &3&=우수)",
        )
        treasure_stats = set_specific_treasures(
            treasure_stats, treasure_data, chapter_id
        )
    save_stats["treasures"] = treasure_stats
    print("성공적으로 보물 설정")
    return save_stats


def specific_stages_all_chapters(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing treasure levels"""

    chapter_ids = story_level_id_selector.select_specific_chapters()

    treasure_stats = save_stats["treasures"]

    stage_ids = story_level_id_selector.select_levels(None)
    treasure_data = [-1] * 48
    for i, chapter_id in enumerate(chapter_ids):
        chapter_id = main_story.format_story_id(chapter_id)
        if i == 0:
            treasure_data = user_input_handler.handle_all_at_once(
                stage_ids,
                True,
                treasure_data,
                list(range(0, 48)),
                "treasure level",
                "stage",
                "(&0&=없음, &1&=열등, &2&=정상, &3&=우수)",
            )
        treasure_stats = set_specific_treasures(
            treasure_stats, treasure_data, chapter_id
        )
    save_stats["treasures"] = treasure_stats
    print("성공적으로 보물 설정")
    return save_stats



"""Handler for editing tower stages"""
from typing import Any

from . import event_stages
from ... import user_input_handler

def edit_tower(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing tower stages"""

    stage_data = save_stats["tower"]["progress"]
    stage_data = {
        "Value": stage_data,
        "Lengths": {"stars": stage_data["stars"], "stages": stage_data["stages"]},
    }

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "타워 ID를 입력합니다(&이벤트 출시 순서 전투 고양이&를 찾아 &이벤트& 및 &장갑?&을 지나 스크롤하여 &타워& ID를 찾습니다)(&all&을 입력하여 범위(예: &1&-&49&, 또는 공백으로 구분된 ID(예: &5 4 7&)): "
        ),
        stage_data["Value"]["total"],
    )
    save_stats["tower"]["progress"] = event_stages.stage_handler(
        stage_data, ids, 0, False
    )["Value"]
    save_stats["tower"]["progress"]["total"] = stage_data["Value"]["total"]
    save_stats["tower"]["progress"]["stars"] = stage_data["Lengths"]["stars"]
    save_stats["tower"]["progress"]["stages"] = stage_data["Lengths"]["stages"]

    return save_stats



from typing import Any, Optional
from ... import user_input_handler, server_handler, helper, adb_handler
from ..levels import clear_tutorial


def select(save_stats: dict[str, Any]) -> dict[str, Any]:
    helper.check_changes(None)
    options = [
        "전송 및 확인 코드를 사용하여 게임에서 저장 데이터 다운로드",
        "파일에서 저장 파일 선택",
        "adb를 사용하여 루팅된 장치에서 저장을 가져옵니다.",
        "json에서 저장 데이터 불러오기",
    ]
    index = (
        user_input_handler.select_single(
            options, title="저장 데이터를 가져오는 옵션을 선택합니다."
        )
        - 1
    )
    save_path = handle_index(index)
    if not save_path:
        return save_stats
    helper.set_save_path(save_path)
    data = helper.load_save_file(save_path)
    save_stats = data["save_stats"]
    if save_path.endswith(".json"):
        input(
            "저장 데이터가 json 형식인 것 같습니다. json 데이터를 로드하려면 json 옵션을 가져오기 위해 사용하십시오.\n계속하려면 엔터 키를 누르십시오...:"
        )
    if not clear_tutorial.is_tutorial_cleared(save_stats):
        save_stats = clear_tutorial.clear_tutorial(save_stats)
    return save_stats


def handle_index(index: int) -> Optional[str]:
    path = None
    if index == 0:
        print("Enter details for data transfer:")
        path = server_handler.download_handler()
    elif index == 1:
        print("Select save file:")
        path = helper.select_file(
            "Select a save file:",
            helper.get_save_file_filetype(),
            initial_file=helper.get_save_path_home(),
        )
    elif index == 2:
        print("Enter details for save pulling:")
        game_versions = adb_handler.find_game_versions()
        if not game_versions:
            game_version = helper.ask_cc()
        else:
            index = (
                user_input_handler.select_single(
                    game_versions, "Select", "Select a game version to pull from:", True
                )
                - 1
            )
            game_version = game_versions[index]
        path = adb_handler.adb_pull_save_data(game_version)
    elif index == 3:
        print("Select save data json file")
        js_path = helper.select_file(
            "Select save data json file",
            [("Json", "*.json")],
            initial_file=helper.get_save_path_home() + ".json",
        )
        if js_path:
            path = helper.load_json_handler(js_path)
    else:
        helper.colored_text("Please enter a recognised option", base=helper.RED)
        return None
    return path



"""Handler for unlocking the enemy guide"""
from typing import Any

from ... import user_input_handler


def enemy_guide(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for unlocking the enemy guide"""

    enemy_guide_stats = save_stats["enemy_guide"]
    total = len(enemy_guide_stats)
    unlock = (
        user_input_handler.colored_input(
            "적 가이드 항목 &(1)&을(를) 제거하거나 &(2)&를 잠금 해제하시겠습니까?:"
        )
        == "2"
    )
    set_val = 1
    if not unlock:
        set_val = 0
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "적 ID를 입력하십시오(ID를 찾으려면 적군 해제 명령 전투 고양이를 조회하십시오)(모든 범위를 얻으려면 &all&을 입력할 수 있습니다(예: &1&-&50& 또는 공백으로 구분된 ID(예: &5 4 7&))."
        ),
        total,
    )

    for enemy_id in ids:
        if enemy_id >= 2:
            enemy_id -= 2
        if enemy_id >= len(enemy_guide_stats):
            print(f"유효하지 않은 적 ID: {enemy_id+2}")
            continue
        enemy_guide_stats[enemy_id] = set_val
    save_stats["enemy_guide"] = enemy_guide_stats
    if not unlock:
        print("적 가이드 항목을 성공적으로 제거했습니다.")
    else:
        print("적 가이드 항목을 성공적으로 잠금 해제했습니다.")
    return save_stats



from . import config_manager, helper
import os


class PropertySet:
    def __init__(self, locale: str, property: str):
        self.locale = locale
        self.path = os.path.join(
            helper.get_local_files_path(), "locales", locale, property + ".properties"
        )
        if not os.path.exists(self.path):
            os.makedirs(os.path.dirname(self.path))
        self.properties: dict[str, str] = {}
        self.parse()

    def parse(self):
        lines = helper.read_file_string(self.path).splitlines()
        for line in lines:
            if line.startswith("#") or line == "":
                continue
            parts = line.split("=")
            if len(parts) < 2:
                continue
            key = parts[0]
            value = "=".join(parts[1:])
            self.properties[key] = value

    def get_key(self, key: str) -> str:
        return self.properties[key].replace("\\n", "\n")

    @staticmethod
    def from_config(property: str) -> "PropertySet":
        return PropertySet(config_manager.get_config_value("LOCALE"), property)


class LocalManager:
    def __init__(self, locale: str):
        self.locale = locale
        self.path = os.path.join(helper.get_local_files_path(), "locales", locale)
        if not os.path.exists(self.path):
            os.makedirs(self.path)
        self.properties: dict[str, PropertySet] = {}
        self.is_en = locale == "en"
        if not self.is_en:
            self.en_path = os.path.join(helper.get_local_files_path(), "locales", "en")
            if not os.path.exists(self.en_path):
                os.makedirs(self.en_path)
            self.en_properties: dict[str, PropertySet] = {}
        self.parse()

    def parse(self):
        for file in helper.get_files_in_dir(self.path):
            file_name = os.path.basename(file)
            if file_name.endswith(".properties"):
                self.properties[file_name[:-11]] = PropertySet(
                    self.locale, file_name[:-11]
                )
        if not self.is_en:
            for file in helper.get_files_in_dir(self.en_path):
                file_name = os.path.basename(file)
                if file_name.endswith(".properties"):
                    self.en_properties[file_name[:-11]] = PropertySet(
                        "en", file_name[:-11]
                    )

    def get_key(self, property: str, key: str) -> str:
        return self.properties[property].get_key(key)

    def search_key(self, key: str) -> str:
        value = None
        for prop in self.properties.values():
            if key in prop.properties:
                value = prop.get_key(key)
        if not self.is_en:
            for prop in self.en_properties.values():
                if key in prop.properties:
                    value = prop.get_key(key)
        if value is None:
            raise KeyError(f"Key {key} not found")

        while True:
            start = value.find("{{")
            if start == -1:
                break
            end = value.find("}}")
            if end == -1:
                break
            value = value.replace(
                value[start : end + 2], self.search_key(value[start + 2 : end])
            )
        return value

    @staticmethod
    def from_config() -> "LocalManager":
        return LocalManager(config_manager.get_config_value("LOCALE"))

    @staticmethod
    def get_locales() -> list[str]:
        return helper.get_dirs(os.path.join(helper.get_local_files_path(), "locales"))



"""Handler for getting the gold pass"""

import datetime
import random
import time
from typing import Any
from ... import helper, user_input_handler


def remove_gold_pass_val(save_stats: dict[str, Any]) -> dict[str, Any]:
    """
    Remove the gold pass

    Args:
        save_stats (dict[str, Any]): The save stats

    Returns:
        dict[str, Any]: The save stats
    """

    gold_pass = save_stats["gold_pass"]

    gold_pass["officer_id"]["Value"] = 0xFFFFFFFF
    gold_pass["renewal_times"]["Value"] = 0
    gold_pass["start_date"] = 0
    gold_pass["expiry_date"] = 0
    gold_pass["unknown_2"][0] = 0
    gold_pass["unknown_2"][1] = 0

    gold_pass["start_date_2"] = 0
    gold_pass["expiry_date_2"] = 0
    gold_pass["unknown_3"] = 0
    gold_pass["flag_2"]["Value"] = 0
    gold_pass["expiry_date_3"] = 0

    gold_pass["unknown_4"]["Value"] = 0
    gold_pass["unknown_5"]["Value"] = 0
    gold_pass["unknown_6"]["Value"] = 0
    save_stats["gold_pass"] = gold_pass
    save_stats["login_bonuses"][5100] = 0

    return save_stats


def get_gold_pass_val(
    save_stats: dict[str, Any], total_days: int, officer_id: int
) -> dict[str, Any]:
    """
    Give the gold pass

    Args:
        save_stats (dict[str, Any]): The save stats
        total_days (int): The total days
        officer_id (int): The officer ID

    Returns:
        dict[str, Any]: The save stats
    """

    gold_pass = save_stats["gold_pass"]

    start_date = int(time.time())
    expiry_date = start_date + datetime.timedelta(days=total_days).total_seconds()
    expiry_date_2 = start_date + datetime.timedelta(days=total_days * 2).total_seconds()

    gold_pass["officer_id"]["Value"] = officer_id
    if gold_pass["renewal_times"]["Value"] == 0:
        gold_pass["renewal_times"]["Value"] = 1
    gold_pass["renewal_times"]["Value"] += 1
    gold_pass["start_date"] = start_date
    gold_pass["expiry_date"] = expiry_date
    gold_pass["unknown_2"][0] = expiry_date
    gold_pass["unknown_2"][1] = expiry_date_2

    gold_pass["start_date_2"] = start_date
    gold_pass["expiry_date_2"] = expiry_date_2
    gold_pass["unknown_3"] = start_date
    gold_pass["flag_2"]["Value"] = 2
    gold_pass["expiry_date_3"] = expiry_date

    gold_pass["unknown_4"]["Value"] = 0
    gold_pass["unknown_5"]["Value"] = 1
    gold_pass["unknown_6"]["Value"] = 0
    save_stats["gold_pass"] = gold_pass
    save_stats["login_bonuses"][5100] = 0

    return save_stats


def get_random_officer_id() -> int:
    """Get a random officer ID"""

    return random.randint(1, 2**16 - 1)


def get_gold_pass(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Give the gold pass"""

    officer_id = user_input_handler.colored_input(
        "원하는 &임원 ID&를 입력하세요.:"
    )
    if officer_id == "":
        officer_id = get_random_officer_id()
    elif officer_id == "-1":
        officer_id = -1
    else:
        officer_id = helper.check_int_max(officer_id)

    if officer_id is None:
        officer_id = 0

    if officer_id == -1:
        save_stats = remove_gold_pass_val(save_stats)
        helper.colored_text("골드 패스를 성공적으로 제거했습니다.", helper.GREEN)
    else:
        helper.colored_text(f"임원 ID: &{officer_id}&", helper.GREEN, helper.WHITE)
        save_stats = get_gold_pass_val(save_stats, 30, officer_id)
        helper.colored_text("성공적으로 골드 패스를 제공했습니다.", helper.GREEN)

    return save_stats



"""Scheme item edit"""
from typing import Any

from ... import csv_handler, game_data_getter, helper, user_input_handler


def get_item_names(is_jp: bool) -> list[str]:
    """Get the item names

    Args:
        is_jp (bool): If the data is for jp

    Returns:
        list[str]: The item names
    """
    item_names = game_data_getter.get_file_latest(
        "resLocal", "GatyaitemName.csv", is_jp
    )
    if item_names is None:
        helper.error_text("Failed to get item names")
        return []

    item_names = csv_handler.parse_csv(
        item_names.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )
    names: list[str] = []
    for item in item_names:
        names.append(item[0])
    return names


def get_scheme_data(is_jp: bool) -> list[list[int]]:
    """Get the scheme data

    Args:
        is_jp (bool): If the data is for jp

    Returns:
        list[list[int]]: The scheme data
    """
    scheme_data = game_data_getter.get_file_latest(
        "DataLocal", "schemeItemData.tsv", is_jp
    )
    if scheme_data is None:
        helper.error_text("체계 데이터를 가져오지 못했습니다.")
        return []

    scheme_data_data = helper.parse_int_list_list(
        csv_handler.parse_csv(
            scheme_data.decode("utf-8"),
            delimeter="\t",
        )
    )
    return scheme_data_data


def get_scheme_names(is_jp: bool, scheme_data: list[list[int]]) -> dict[int, str]:
    """Get the scheme names"""

    file_data = game_data_getter.get_file_latest("resLocal", "localizable.tsv", is_jp)
    if file_data is None:
        helper.error_text("체계 이름을 가져오지 못했습니다.")
        return {}

    localizable = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter="\t",
    )
    names: dict[int, str] = {}
    for scheme in scheme_data[1:]:
        scheme_id = scheme[0]
        for name in localizable:
            scheme_str = f"scheme_popup_{scheme_id}"
            if name[0] == scheme_str:
                scheme_name = name[1].replace("<flash>", "").replace("</flash>", "")
                names[scheme_id] = scheme_name
                break
    return names


def get_cat_name(cat_id: int, is_jp: bool, cc: str) -> str:
    """Get the cat name"""

    file_data = game_data_getter.get_file_latest(
        "resLocal", f"Unit_Explanation{cat_id+1}_{cc}.csv", is_jp
    )
    if file_data is None:
        helper.error_text("Failed to get cat names")
        return ""

    cat_name = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )
    return cat_name[0][0]


def edit_scheme_data(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing scheme data"""

    is_jp = helper.check_data_is_jp(save_stats)
    data = get_scheme_data(is_jp)
    names = get_scheme_names(is_jp, data)
    item_names = get_item_names(is_jp)

    options: list[str] = []
    for scheme in data[1:]:
        scheme_id = scheme[0]
        is_cat = scheme[2] == 1
        item_id = scheme[3]
        amount = scheme[4]
        try:
            scheme_name = names[scheme_id]
        except KeyError:
            continue
        string = "\n\t"
        if is_cat:
            cat_name = get_cat_name(item_id, is_jp, helper.get_lang(is_jp))
            string += scheme_name.replace("%@", cat_name)
        else:
            try:
                item_name = item_names[item_id]
            except IndexError:
                continue
            string += scheme_name
            first_index = string.find("%@")
            second_index = string.find("%@", first_index + 1)
            string = (
                string[:first_index]
                + str(amount)
                + " "
                + item_name
                + string[second_index + 2 :]
            )

        string = string.replace("<br>", "\n\t")
        options.append(string)

    scheme_ids = user_input_handler.select_not_inc(options, "get")
    scheme_data = save_stats["item_schemes"]
    for scheme_index in scheme_ids:
        try:
            scheme_id = data[scheme_index + 1][0]
        except IndexError:
            continue
        obtain_ids: list[int] = scheme_data["to_obtain_ids"]
        obtain_ids.append(scheme_id)
        received_ids: list[int] = scheme_data["received_ids"]
        if scheme_id in received_ids:
            received_ids.remove(scheme_id)
        scheme_data["to_obtain_ids"] = obtain_ids
        scheme_data["received_ids"] = received_ids
    save_stats["item_schemes"] = scheme_data
    return save_stats



"""Handler for editting uncanny legends"""
from typing import Any

from . import event_stages
from ... import user_input_handler

def edit_uncanny(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting uncanny legends"""
    stage_data = save_stats["uncanny"]
    lengths = stage_data["Lengths"]

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "단계 ID 입력(예: &1& = 새 범례, &2& = 여기 드래곤)(모든 범위를 가져오려면 &all&을 입력할 수 있습니다. 예: &1&-&49& 또는 공백으로 구분된 ID(예: &5 4 7&):"
        ),
        lengths["total"],
    )
    save_stats["uncanny"] = event_stages.stage_handler(stage_data, ids, -1)

    return save_stats

def is_ancient_curse_clear(save_stats: dict[str, Any]) -> bool:
    """
    Check if the ancient curse is cleared

    Args:
        save_stats (dict[str, Any]): The save stats

    Returns:
        bool: If the ancient curse is cleared
    """
    return save_stats["uncanny"]["Value"]["clear_progress"][0][0] >= 1



"""Handler for selecting and running editor features"""

from typing import Any, Union

from . import (
    helper,
    user_input_handler,
    config_manager,
)
from .edits import basic, cats, gamototo, levels, other, save_management


def fix_elsewhere_old(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Fix the elsewhere error using 2 save files"""

    main_token = save_stats["token"]
    main_iq = save_stats["inquiry_code"]
    input(
        "다른곳에 오류가 없고 금지되지 않은 현재 게임 내에서 로드된 저장 파일을 선택하십시오.\n계속하려면 엔터 키를 누르십시오."
    )
    new_path = helper.select_file(
        "클리어 세이브 파일 선택",
        helper.get_save_file_filetype(),
        helper.get_save_path(),
    )
    if not new_path:
        print("세이브 파일을 선택해주세요")
        return save_stats

    data = helper.load_save_file(new_path)
    new_stats = data["save_stats"]
    new_token = new_stats["token"]
    new_iq = new_stats["inquiry_code"]
    save_stats["token"] = new_token
    save_stats["inquiry_code"] = new_iq

    helper.colored_text(f"대체된 문의 코드: &{main_iq}& with &{new_iq}&")
    helper.colored_text(f"교체된 토큰: &{main_token}& with &{new_token}&")
    return save_stats


FEATURES: dict[str, Any] = {
    "저장 관리": {
        "저장하기": save_management.save.save_save,
        "변경 사항 저장 및 게임 서버에 업로드(전송 및 확인 코드 받기)": save_management.server_upload.save_and_upload,
        "변경 사항을 파일에 저장": save_management.save.save,
        "adb를 사용하여 변경 사항을 저장하고 저장 데이터를 게임에 푸시합니다(게임을 다시 열지 않음).": save_management.save.save_and_push,
        "adb를 사용하여 변경 사항을 저장하고 저장 데이터를 게임에 푸시(게임 다시 열기)": save_management.save.save_and_push_rerun,
        "저장 데이터를 json으로 내보내기": save_management.other.export,
        "adb로 저장 데이터 지우기(게임을 재설치하지 않고 새 계정을 생성하는 데 사용)": save_management.other.clear_data,
        "추적 금지 항목 업로드(저장 또는 종료 시 자동으로 수행됨)": save_management.server_upload.upload_metadata,
        "세이브 데이터 불러오기": save_management.load.select,
        "저장 데이터를 다른 버전으로 변환": save_management.convert.convert_save,
        # "Manage Presets": preset_handler.preset_manager,
    },
    "아이템류": {
        "통조림": basic.basic_items.edit_cat_food,
        "경험치": basic.basic_items.edit_xp,
        "티켓류": {
            "노말 티켓": basic.basic_items.edit_normal_tickets,
            "레어 티켓": basic.basic_items.edit_rare_tickets,
            "플레티넘 티켓": basic.basic_items.edit_platinum_tickets,
            "플레티넘 조각": basic.basic_items.edit_platinum_shards,
            "레전드 티켓": basic.basic_items.edit_legend_tickets,
        },
        "NP": basic.basic_items.edit_np,
        "리더쉽": basic.basic_items.edit_leadership,
        "배틀 아이템": basic.basic_items.edit_battle_items,
        "캣츠아이": basic.catseyes.edit_catseyes,
        "개다래 / 베히모스 스톤": basic.catfruit.edit_catfruit,
        "본능 구슬": basic.talent_orbs_new.edit_talent_orbs,
        "가마토토 드링크": basic.basic_items.edit_catamins,
        "아이템구성 (금지할 수 없는 아이템을 얻을 수 있음)": other.scheme_item.edit_scheme_data,
    },
    "가마토토 / 오토토": {
        "오토토 조수": basic.basic_items.edit_engineers,
        "오토토 재료": basic.ototo_base_mats.edit_base_mats,
        "가마토토 드링크": basic.basic_items.edit_catamins,
        "가마토토 경험치 / 레벨": gamototo.gamatoto_xp.edit_gamatoto_xp,
        "오토토 대포": gamototo.ototo_cat_cannon.edit_cat_cannon,
        "가마토토 도우미": gamototo.helpers.edit_helpers,
        "충돌 하는 가마토토 오류 수정하기": gamototo.fix_gamatoto.fix_gamatoto,
    },
    "고양이 / 스페셜 스킬[지갑]": {
        "고양이를 추가 / 고양이를 제거 ": {
            "➕고양이추가": cats.get_remove_cats.get_cat,
            "➖고양이제거": cats.get_remove_cats.remove_cats,
        },
        "고양이 업그레이드": cats.upgrade_cats.upgrade_cats,
        "고양이 진화시키기[3진]": {
            "고양이 3진화 추가": cats.evolve_cats.get_evolve,
            "고양이 3진화 제거": cats.evolve_cats.remove_evolve,
            "강제로 진정한 형태의 고양이[냥코id코드로 추가]\n[⚠️모든 고양이 선택 금지] (진정한 형태가 없는 캐릭터를 위한 빈 캐릭터로 이어질 것입니다.)": cats.evolve_cats.get_evolve_forced,
        },
        "본능": {
            "선택한 각 고양이의 본능을 개별적으로 설정": cats.talents.edit_talents_individual,
            "최대 / 선택된 모든 고양이 본능 제거": cats.talents.max_all_talents,
        },
        "고양이 가이드 수집/제거": {
            "고양이 가이드 항목 설정(참조하지 않음)": cats.clear_cat_guide.collect_cat_guide,
            "고양이 가이드 항목 청구 취소": cats.clear_cat_guide.remove_cat_guide,
        },
        ##'스테이지 단위 드랍을 가져오면 이 스테이지를 클리어하여 특별한 고양이 대화 상자를 제거합니다[안씀]': cats.chara_drop.get_character_drops,
        "특수 기술/능력 업그레이드": cats.upgrade_blue.upgrade_blue,
    },
    "레벨 / 보물": {
        "메인 스토리 챕터 클리어/언클리어": {
            "선택한 모든 챕터의 모든 챕터에서 각 스테이지 클리어": levels.main_story.clear_all,
            "선택한 각 챕터의 모든 챕터에서 각 스테이지 클리어": levels.main_story.clear_each,
        },
        "보물": {
            "트레저 그룹 (예: 에너지 드링크, 아쿠아 크리스탈 등)": levels.treasures.treasure_groups,
            "개별적으로 특정 단계 및 특정 챕터": levels.treasures.specific_stages,
            "특정 스테이지와 챕터를 한 번에": levels.treasures.specific_stages_all_chapters,
        },
        "좀비스테이지 / 노출": levels.outbreaks.edit_outbreaks,
        "이벤트 스테이지": levels.event_stages.event_stages,
        "레전드 스토리": levels.event_stages.stories_of_legend,
        "신 레전드스토리": levels.uncanny.edit_uncanny,
        "마계편 스테이지/게이트 클리어": levels.aku.edit_aku,
        "마계편 스테이지/게이트 잠금 해제": levels.unlock_aku_realm.unlock_aku_realm,
        ##"Gauntlets": levels.gauntlet.edit_gauntlet,
        ##"Collab Gauntlets": levels.gauntlet.edit_collab_gauntlet,
        "타워": levels.towers.edit_tower,
        "레전드리스토리[빅 하앜마양 강림 클리어]": levels.behemoth_culling.edit_behemoth_culling,
        "미래편 시간 초과 점수": levels.itf_timed_scores.timed_scores,
        "도전 전투 점수": basic.basic_items.edit_challenge_battle,
        "튜토리얼 클리어": levels.clear_tutorial.clear_tutorial,
        "랭킹 점수(입문자의 전당)": basic.basic_items.edit_dojo_score,
        "수수께끼 단계 추가": levels.enigma_stages.edit_enigma_stages,
        "필리버스터 단계 언클리어": levels.allow_filibuster_clearing.allow_filibuster_clearing,
        "레전드 퀘스트": levels.legend_quest.edit_legend_quest,
    },
    "계정 / 문의 코드 / 토큰 ": {
        "문의 코드": basic.basic_items.edit_inquiry_code,
        "토큰": basic.basic_items.edit_token,
        "다른 곳 오류 수정 / 계정 차단 해제": other.fix_elsewhere.fix_elsewhere,
        ##"이전 수정 다른 곳 오류 / 계정 금지 해제(2개의 저장 파일 필요)": fix_elsewhere_old,
        "새로운 조회 코드 및 토큰 생성": other.create_new_account.create_new_account,
    },
    "기타": {
        "레어 가챠 씨앗": basic.basic_items.edit_rare_gacha_seed,
        "잠금 해제된 장비 슬롯": basic.basic_items.edit_unlocked_slots,
        "재시작 팩 받기/ 귀환자 모드\n[⚠️재시작팩 하실려면 먼저 하셔야됩니다\n아니면초기화됩니다]": basic.basic_items.edit_restart_pack,
        "냥코 메달": other.meow_medals.medals,
        "플레이 타임": other.play_time.edit_play_time,
        "적 가이드 항목 잠금 해제/제거": other.unlock_enemy_guide.enemy_guide,
        "개다레 챌린지 / 미션": other.missions.edit_missions,
        "일반 티켓 최대 거래 진행률(금지할 수 없는 희귀 티켓 허용)": other.trade_progress.set_trade_progress,
        "골드 패스 받기/제거": other.get_gold_pass.get_gold_pass,
        "클레임 / 모든 사용자 등급 보상 제거(아무 아이템도 주지 않음)": other.claim_user_rank_rewards.edit_rewards,
        "고양이 신전 레벨 / 경험치[모르겟네요]": other.cat_shrine.edit_shrine_xp,
    },
    "수정": {
        "시간 오류 수정": other.fix_time_issues.fix_time_issues,
        "장비 메뉴 잠금 해제": other.unlock_equip_menu.unlock_equip,
        "튜토리얼 클리어": levels.clear_tutorial.clear_tutorial,
        "다른 곳 오류 수정 / 계정 차단 해제": other.fix_elsewhere.fix_elsewhere,
        ##"이전 수정 다른 곳 오류 / 계정 금지 해제(2개의 저장 파일 필요)": fix_elsewhere_old,
        "게임 충돌에서 가마토토 수정": gamototo.fix_gamatoto.fix_gamatoto,
    },
    "에디터 편집[사용하지마세요]": {
        "Edit LOCALIZATION": config_manager.edit_locale,
        "Edit DEFAULT_COUNTRY_CODE": config_manager.edit_default_gv,
        "Edit DEFAULT_SAVE_PATH": config_manager.edit_default_save_file_path,
        "Edit FIXED_SAVE_PATH": config_manager.edit_fixed_save_path,
        "Edit EDITOR settings": config_manager.edit_editor_settings,
        "Edit START_UP settings": config_manager.edit_start_up_settings,
        "Edit SAVE_CHANGES settings": config_manager.edit_save_changes_settings,
        "Edit SERVER settings": config_manager.edit_server_settings,
        "Edit config path": config_manager.edit_config_path,
    },
    "종료": helper.exit_check_changes,
}


def get_feature(
    selected_features: Any, search_string: str, results: dict[str, Any]
) -> dict[str, Any]:
    """Search for a feature if the feature name contains the search string"""

    for feature in selected_features:
        feature_data = selected_features[feature]
        if isinstance(feature_data, dict):
            feature_data = get_feature(feature_data, search_string, results)
        if search_string.lower().replace(" ", "") in feature.lower().replace(" ", ""):
            results[feature] = selected_features[feature]
    return results


def show_options(
    save_stats: dict[str, Any], features_to_use: dict[str, Any]
) -> dict[str, Any]:
    """Allow the user to either enter a feature number or a feature name, and get the features that match"""

    if (
        not config_manager.get_config_value_category("EDITOR", "SHOW_CATEGORIES")
        and FEATURES == features_to_use
    ):
        user_input = ""
    else:
        prompt = (
            "무엇을 편집하시겠습니까(일부 옵션에는 다른 기능이 포함되어 있음)"
        )
        if config_manager.get_config_value_category(
            "EDITOR", "SHOW_FEATURE_SELECT_EXPLANATION"
        ):
            prompt += "\n숫자를 입력하여 기능을 실행하거나 단어를 입력하여 해당 기능을 검색할 수 있습니다(예: catfood를 입력하면 통조림 기능이 실행되고 티켓을 입력하면 티켓을 편집하는 모든 기능이 표시됨)\n엔터를 눌러 목록을 볼 수 있습니다. 모든 기능의"
        user_input = user_input_handler.colored_input(f"{prompt}:\n")
    user_int = helper.check_int(user_input)
    results = []
    if user_int is None:
        results = get_feature(features_to_use, user_input, {})
    else:
        if user_int < 1 or user_int > len(features_to_use) + 1:
            helper.colored_text("범위를 벗어난 값", helper.RED)
            return show_options(save_stats, features_to_use)
        if FEATURES != features_to_use:
            if user_int - 2 < 0:
                return menu(save_stats)
            results = features_to_use[list(features_to_use)[user_int - 2]]
        else:
            results = features_to_use[list(features_to_use)[user_int - 1]]
    if not isinstance(results, dict):
        save_stats_return = results(save_stats)
        if save_stats_return is None:
            return save_stats
        return save_stats_return
    if len(results) == 0:
        helper.colored_text("해당 이름의 기능을 찾을 수 없습니다.", helper.RED)
        return menu(save_stats)
    if len(results) == 1 and isinstance(list(results.values())[0], dict):
        results = results[list(results)[0]]
    if len(results) == 1:
        save_stats_return = results[list(results)[0]](save_stats)
        if save_stats_return is None:
            return save_stats
        return save_stats_return

    helper.colored_list(["뒤로가기"] + list(results))
    return show_options(save_stats, results)


def menu(
    save_stats: dict[str, Any], path_save: Union[str, None] = None
) -> dict[str, Any]:
    """Show the menu and allow the user to select a feature to edit"""

    if path_save:
        helper.set_save_path(path_save)
    if config_manager.get_config_value_category("EDITOR", "SHOW_CATEGORIES"):
        helper.colored_list(list(FEATURES))
    save_stats = show_options(save_stats, FEATURES)

    return save_stats



"""Fix gamatoto from crashing the game"""
from typing import Any


def fix_gamatoto(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Fix gamatoto from crashing the game"""

    save_stats["gamatoto_skin"]["Value"] = 2
    print("게임 충돌로부터 가마토토를 성공적으로 수정했습니다.")
    return save_stats



"""Handler for selecting story levels"""
from typing import Optional

from ... import user_input_handler, helper
from . import main_story


def select_specific_chapters() -> list[int]:
    """Select specific levels"""

    print("어떤 챕터를 선택하시겠습니까?")
    ids = user_input_handler.select_not_inc(main_story.CHAPTERS, "clear")
    return ids


def get_option():
    """Get option"""

    options = [
        "단계 ID로 특정 수준 선택",
        "특정 단계까지의 모든 레벨 선택",
        "모든 레벨 선택",
    ]
    return user_input_handler.select_single(options)


def select_levels(
    chapter_id: Optional[int], forced_option: Optional[int] = None, total: int = 48
) -> list[int]:
    """Select levels"""

    if forced_option is None:
        choice = get_option()
    else:
        choice = forced_option
    if choice == 1:
        return select_specific_levels(chapter_id, total)
    if choice == 2:
        return select_levels_up_to(chapter_id, total)
    if choice == 3:
        return select_all(total)
    return []


def select_specific_levels(chapter_id: Optional[int], total: int) -> list[int]:
    """Select specific levels"""

    print("어떤 수준을 선택하시겠습니까?")
    if chapter_id is not None:
        helper.colored_text(
            f"Chapter: &{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    ids = user_input_handler.get_range_ids(
        "레벨 ID(예: &1&=한국, &2&=몽골)", total
    )
    ids = helper.check_clamp(ids, total, 1, -1)
    return ids


def select_levels_up_to(chapter_id: Optional[int], total: int) -> list[int]:
    """Select levels up to a certain level"""

    print("어떤 수준을 선택하시겠습니까?")
    if chapter_id is not None:
        helper.colored_text(
            f"Chapter: &{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    stage_id = user_input_handler.get_int(
        f"원하는단계 단계 ID를 최대(및 포함)까지 입력하세요(예: &1&=한국 클리어, &2&=한국 &또는& 몽골 클리어, &{total}&=all)?:"
    )
    stage_id = helper.clamp(stage_id, 1, total)
    return list(range(0, stage_id))


def select_all(total: int) -> list[int]:
    """Select all levels"""

    return list(range(0, total))


def select_level_progress(
    chapter_id: Optional[int], total: int, examples: Optional[list[str]] = None
) -> int:
    """Select level progress"""

    if examples is None:
        examples = [
            "한국",
            "몽골",
        ]

    print("어떤 레벨까지 클리어하고 싶나요?")
    if chapter_id is not None:
        helper.colored_text(
            f"챕터: &{chapter_id+1}& : &{main_story.CHAPTERS[chapter_id]}&"
        )
    progress = user_input_handler.get_int(
        f"원하는 단계 ID를 입력하세요.클리어/언클리어 (e.g &1&={examples[0]} 클리어, &2&={examples[0]} &또는& {examples[1]} 클리어, &{total}&=all, &0&=언클리어 all)?:"
    )
    progress = helper.clamp(progress, 0, total)
    return progress



"""Handler for selecting cat ids"""

import os
from multiprocessing import Process
from typing import Any, Callable, Optional


from ... import (
    csv_handler,
    game_data_getter,
    helper,
    user_input_handler,
)
from ..levels import treasures
from . import cat_helper


def select_cats(save_stats: dict[str, Any], current: bool = True) -> list[int]:
    """Select cats"""

    options: dict[str, Callable[[dict[str, Any]], list[int]]] = {
        "현재 잠금 해제된 고양이 선택": select_current_cats,
        "특정 희귀도의 고양이 선택": select_cats_rarity,
        "특정 고양이 ID 선택": select_cats_range,
        "특정 뽑기 배너의 고양이 선택": select_cats_gatya_banner,
        "모든 고양이 선택": get_all_cats,
        "고양이 이름으로 검색": select_cat_names,
        "획득 가능한 모든 고양이 선택": select_cats_obtainable,
    }
    if not current:
        del options["현재 잠금 해제된 고양이 선택"]

    choice_index = (
        user_input_handler.select_single(list(options.keys()), title="고양이 선택:") - 1
    )
    cat_ids = options[list(options)[choice_index]](save_stats)
    return cat_ids


def select_cats_obtainable(save_stats: dict[str, Any]) -> list[int]:
    """
    Select cats that can be obtained

    Args:
        save_stats (dict[str, Any]): Save stats

    Returns:
        list[int]: Cat ids
    """
    return filter_obtainable_cats(save_stats, get_all_cats(save_stats))


def select_current_cats(save_stats: dict[str, Any]) -> list[int]:
    """Select current cats"""

    cats = save_stats["cats"]
    cat_ids: list[int] = []
    for i, cat_val in enumerate(cats):
        if cat_val == 1:
            cat_ids.append(i)
    return cat_ids


def select_cats_rarity(save_stats: dict[str, Any]) -> list[int]:
    """Select cats of a certain rarity"""

    ids = user_input_handler.select_not_inc(
        options=cat_helper.TYPES,
        mode="select",
    )
    is_jp = helper.is_jp(save_stats)

    cat_ids = cat_helper.get_rarity(ids, is_jp)
    return cat_ids


def select_cats_range(save_stats: dict[str, Any]) -> list[int]:
    """Select cats in a range"""

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "고양이 ID를 입력하십시오(ID를 찾으려면 냥코id사이트.kro.kr에서 조회하십시오)(&all&을 입력하여 범위(예: &1&-&50& 또는 공백으로 구분된 ID(예: &5 4 7&)를 얻을 수 있음):"
        ),
        length=len(save_stats["cats"]),
    )
    return ids


def select_cats_gatya_banner(save_stats: dict[str, Any]) -> list[int]:
    """Select cats for a specific gacha banner"""
    is_jp = helper.is_jp(save_stats)
    file_data = game_data_getter.get_file_latest(
        "DataLocal", "GatyaDataSetR1.csv", is_jp
    )
    if file_data is None:
        helper.colored_text("Failed to get gatya banners")
        return []
    data = helper.parse_int_list_list(csv_handler.parse_csv(file_data.decode("utf-8")))
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "가챠배너 아이디 입력 (원하는 가챠배너 조회 후 상단의 이미지 클릭 후 파일명 마지막 자리 찾기 (예. royal fest = 602))(&all& 입력하시면 모두 받으실 수 있습니다. 범위(예: &1&-&50&) 또는 공백으로 구분된 ID(예: &5 4 7&):"
        ),
        length=len(data),
    )
    data = treasures.remove_negative_1(data)
    cat_ids: list[int] = []
    for c_id in ids:
        cat_ids.extend(data[c_id])
    return cat_ids


def get_all_cats(save_stats: dict[str, Any]) -> list[int]:
    """Get all cats"""

    return list(range(len(save_stats["cats"])))


def select_cat_names(save_stats: dict[str, Any]) -> list[int]:
    """
    select_cat_names

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[int]: cat ids
    """
    all_names = get_cat_names(save_stats)
    if all_names is None:
        return []
    name = user_input_handler.colored_input("고양이 이름 입력[영어로]:")
    found_names = search_cat_names(name, all_names)
    found_names = filter_cat_names(found_names)
    if not found_names:
        print("해당 이름을 가진 고양이를 찾을 수 없습니다.")
        return []

    cat_ids: list[int] = []
    cat_ids_str: list[str] = []
    cat_names: list[str] = []
    for cat_name, cat_id, _ in found_names:
        cat_ids.append(cat_id)
        cat_name = cat_name.replace("&", "\\&")
        cat_names.append(cat_name)
        cat_ids_str.append(f"고양이 아이디: &{cat_id}&")

    print("선택할 고양이의 인덱스 선택(고양이 ID 자체가 아님):")
    indexes = user_input_handler.select_not_inc(
        cat_names, mode="select", extra_data=cat_ids_str
    )
    selected_ids: list[int] = []
    for index in indexes:
        try:
            selected_ids.append(cat_ids[index])
        except IndexError:
            helper.colored_text(
                f"옵션이 너무 높음: {index} -고양이 ID가 아닌 왼쪽의 인덱스를 선택해야 합니다.",
                helper.RED,
            )
    return selected_ids


def get_cat_by_form_and_id(
    all_names: list[tuple[str, int, int]], cat_id: int, form_id: int
) -> Optional[tuple[str, int, int]]:
    """
    Get cat by form and id

    Args:
        all_names (list[tuple[str, int, int]]): all names
        cat_id (int): cat id
        form_id (int): form id

    Returns:
        Optional[tuple[str, int, int]]: cat data
    """
    for cat in all_names:
        if cat[1] == cat_id and cat[2] == form_id:
            return cat
    return None


def get_cat_by_id(
    cat_names: list[tuple[str, int, int]], cat_id_to_search: int
) -> list[tuple[str, int, int]]:
    """
    Get cat by id

    Args:
        cat_names (list[tuple[str, int, int]]): list of cat names
        cat_id_to_search (int): cat id to search for

    Returns:
        Optional[tuple[str, int, int]]: cat name, cat id, cat form
    """
    cats: list[tuple[str, int, int]] = []
    for cat_name, cat_id, cat_form in cat_names:
        if cat_id == cat_id_to_search:
            cats.append((cat_name, cat_id, cat_form))
    return cats


def filter_cat_names(
    cat_names: list[tuple[str, int, int]]
) -> list[tuple[str, int, int]]:
    """
    Filter cat names by only selecting one of the forms

    Args:
        cat_names (list[tuple[str, int, int]]): list of cat names

    Returns:
        list[tuple[str, int, int]]: filtered cat names
    """
    filtered_cat_ids: list[int] = []
    cat_data: list[tuple[str, int, int]] = []
    for cat_name, cat_id, cat_form in cat_names:
        if cat_id not in filtered_cat_ids:
            filtered_cat_ids.append(cat_id)
            cat_data.append((cat_name, cat_id, cat_form))

    return cat_data


def search_cat_names(
    name: str, cat_names: list[tuple[str, int, int]]
) -> list[tuple[str, int, int]]:
    """
    Search cat names

    Args:
        name (str): name to search for
        cat_names (list[tuple[str, int, int]]): list of cat names

    Returns:
        list[tuple[str, int, int]]: list of cat names that match the search
    """

    found_names: list[tuple[str, int, int]] = []

    for cat_name, cat_id, form_id in cat_names:
        if name.lower().replace(" ", "") in cat_name.lower().replace(" ", ""):
            found_names.append((cat_name, cat_id, form_id))
    return found_names


def download_10_files(game_version: str, file_names: list[str]) -> None:
    """
    Download 10 files

    Args:
        game_version (str): game version
        file_names (list[str]): file names
    """
    for file_name in file_names:
        game_data_getter.download_file(game_version, "resLocal", file_name, False)


def get_cat_names(save_stats: dict[str, Any]) -> Optional[list[tuple[str, int, int]]]:
    """
    Get cat names and ids

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        Optional[list[tuple[str, int, int]]]: cat names and ids
    """

    is_jp = helper.is_jp(save_stats)

    path = game_data_getter.get_path("resLocal", "", is_jp)
    if path is None:
        helper.colored_text("고양이 이름을 가져오지 못했습니다.", helper.RED)
        return None
    file_path_dir = os.path.dirname(helper.get_file(path))
    helper.create_dirs(file_path_dir)
    if len(helper.find_files_in_dir(file_path_dir, "Unit_Explanation")) < len(
        save_stats["cats"]
    ):
        helper.colored_text(
            "처음으로 고양이 이름을 다운로드하는 중... (시간이 좀 걸릴 수 있지만 다음에는 훨씬 더 빠를 것입니다)",
            helper.GREEN,
        )
        funcs: list[Process] = []
        version = game_data_getter.get_latest_version(is_jp)
        if version is None:
            helper.colored_text("고양이 이름을 가져오지 못했습니다.", helper.RED)
            return None
        all_file_names: list[str] = []
        for cat_id, _ in enumerate(save_stats["cats"]):
            file_name = f"Unit_Explanation{cat_id+1}_{helper.get_lang(is_jp)}.csv"
            all_file_names.append(file_name)
        file_names_split = helper.chunks(all_file_names, 10)
        for file_names in file_names_split:
            funcs.append(
                Process(
                    target=download_10_files,
                    args=(version, file_names),
                )
            )
        helper.run_in_parallel(funcs)

    names: list[tuple[str, int, int]] = []
    for cat_id, _ in enumerate(save_stats["cats"]):
        file_path = os.path.join(
            file_path_dir, f"Unit_Explanation{cat_id+1}_{helper.get_lang(is_jp)}.csv"
        )
        data = csv_handler.parse_csv(
            helper.read_file_string(file_path),
            delimeter=helper.get_text_splitter(is_jp),
        )
        for form_id, form in enumerate(data):
            name = form[0]
            names.append((name, cat_id, form_id))
    return names


def get_obtainability(save_stats: dict[str, Any]) -> list[int]:
    """
    Get obtainability of cats

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[int]: obtainability of cats (0 = not obtainable, 1 = obtainable)
    """
    file_data = game_data_getter.get_file_latest(
        "DataLocal", "nyankoPictureBookData.csv", helper.is_jp(save_stats)
    )
    if file_data is None:
        helper.colored_text("획득 가능성을 얻지 못했습니다.", helper.RED)
        return []
    data = helper.parse_int_list_list(csv_handler.parse_csv(file_data.decode("utf-8")))
    is_obtainable = helper.copy_first_n(data, 0)
    return is_obtainable


def get_obtainable_cats(save_stats: dict[str, Any]) -> list[int]:
    """
    Get obtainable cats

    Args:
        save_stats (dict[str, Any]): save stats

    Returns:
        list[int]: obtainable cats
    """
    obtainability = get_obtainability(save_stats)
    return [i for i, x in enumerate(obtainability) if x == 1]


def filter_obtainable_cats(save_stats: dict[str, Any], cat_ids: list[int]) -> list[int]:
    """
    Filter obtainable cats in a list of cat ids

    Args:
        save_stats (dict[str, Any]): save stats
        cat_ids (list[int]): cat ids

    Returns:
        list[int]: obtainable cats
    """
    obtainable_cats = get_obtainable_cats(save_stats)
    return [i for i in cat_ids if i in obtainable_cats]



"""Handler for editting meow medals"""
from enum import Enum
import json
from typing import Any, Optional

from ... import helper, user_input_handler, game_data_getter


def get_medal_names(is_jp: bool) -> Optional[list[str]]:
    """Get all medal names"""

    file_data = game_data_getter.get_file_latest("resLocal", "medalname.tsv", is_jp)
    if file_data is None:
        helper.error_text("메달 이름을 가져오지 못했습니다.")
        return None
    medal_names = file_data.decode("utf-8").splitlines()
    names: list[str] = []
    for line in medal_names:
        line_split = line.split("\t")
        name = (
            line_split[0]
            .rstrip("\n")
            .replace("&", "and")
            .replace("â˜…", "")
            .lstrip(" ")
        )
        names.append(name)
    return names


def set_medals(medal_stats: dict[str, Any], ids: list[int]) -> dict[str, Any]:
    """Set the medal stats of a set of medals"""

    for medal_id in ids:
        if medal_id == 0:
            continue
        medal_id -= 1
        if medal_id not in medal_stats["medal_data_1"]:
            if medal_id not in medal_stats["medal_data_2"]:
                medal_stats["medal_data_1"].append(medal_id)
            medal_stats["medal_data_2"][medal_id] = 0
    return medal_stats


def remove_medals(medal_stats: dict[str, Any], ids: list[int]) -> dict[str, Any]:
    """Remove the medal stats of a set of medals"""

    for medal_id in ids:
        if medal_id == 0:
            continue
        medal_id -= 1
        if medal_id in medal_stats["medal_data_1"]:
            medal_stats["medal_data_1"].remove(medal_id)
        if medal_id in medal_stats["medal_data_2"]:
            medal_stats["medal_data_2"].pop(medal_id)
    return medal_stats


class BaseMapIds(Enum):
    """Base map IDs"""

    STORY_CHAPTERS = 3000
    OUTBREAKS_EOC = 20000
    OUTBREAKS_ITF = 21000
    OUTBREAKS_COTC = 22000
    FILIBUSTER = 23000
    LEGEND_STAGES = 0
    EVENT_STAGES = 1000
    TOWER_STAGES = 7000
    LEGEND_QUEST = 16000
    IDI_RE = 4026
    AKU_REALM = 4042
    GAUNTLETS = 24000


class ActionTypes(Enum):
    """Action types"""

    EARN_CENT = 0
    GAMATOTO_EXPLORE = 1
    CAT_BASE_WEAPONS = 2
    USER_RANK = 3
    RECRUIT_GAMATOTO_ASSISTANT = 4


class Medal:
    """Medal"""

    def __init__(self, medal_id: int, grade: int, line: int):
        self.medal_id = medal_id
        self.grade = grade
        self.line = line


class StageMedal(Medal):
    """Stage medal"""

    def __init__(
        self,
        medal_id: int,
        grade: int,
        line: int,
        maps: Optional[list[int]],
        condition: Optional[dict[str, Any]] = None,
        star: Optional[int] = None,
    ):
        super().__init__(medal_id, grade, line)
        self.maps = maps
        self.condition = condition
        self.star = star


class TreasureMedal(StageMedal):
    """Treasure medal"""

    def __init__(
        self,
        medal_id: int,
        grade: int,
        line: int,
        maps: Optional[list[int]],
        treasure: int,
        condition: Optional[dict[str, Any]] = None,
    ):
        super().__init__(medal_id, grade, line, maps, condition)
        self.treasure = treasure


class ActionMedal(Medal):
    """Action medal"""

    def __init__(self, medal_id: int, grade: int, line: int, action: ActionTypes):
        super().__init__(medal_id, grade, line)
        self.action = action


class CharacterMedal(StageMedal):
    """Character medal"""

    def __init__(
        self,
        medal_id: int,
        grade: int,
        line: int,
        maps: Optional[list[int]],
        chara: int,
        condition: Optional[dict[str, Any]] = None,
    ):
        super().__init__(medal_id, grade, line, maps, condition)
        self.chara = chara


class Medals:
    """Medals"""

    def __init__(
        self,
        treasures: list[TreasureMedal],
        characters: list[CharacterMedal],
        actions: list[ActionMedal],
        stages: list[StageMedal],
    ):
        self.treasures = treasures
        self.characters = characters
        self.actions = actions
        self.stages = stages


def get_medal_data(is_jp: bool) -> Optional[Medals]:
    """Get the medal data"""

    file_data = game_data_getter.get_file_latest("DataLocal", "medallist.json", is_jp)
    if file_data is None:
        helper.error_text("메달 데이터를 가져오지 못했습니다.")
        return None
    medal_data = json.loads(file_data.decode("utf-8"))["iconID"]

    treasures: list[TreasureMedal] = []
    characters: list[CharacterMedal] = []
    actions: list[ActionMedal] = []
    stages: list[StageMedal] = []

    for i, medal in enumerate(medal_data):
        if "condition" not in medal:
            medal["condition"] = None
        if "treasure" in medal:
            treasures.append(
                TreasureMedal(
                    i,
                    medal["grade"],
                    medal["line"],
                    medal["map"],
                    medal["treasure"],
                    medal["condition"],
                )
            )
        elif "chara" in medal:
            characters.append(
                CharacterMedal(
                    i,
                    medal["grade"],
                    medal["line"],
                    None,
                    medal["chara"],
                    medal["condition"],
                )
            )
        elif "action" in medal:
            actions.append(
                ActionMedal(
                    i,
                    medal["grade"],
                    medal["line"],
                    ActionTypes(medal["action"]),
                )
            )
        else:
            if "star" not in medal:
                medal["star"] = None
            if "map" not in medal:
                medal["map"] = None
            stages.append(
                StageMedal(
                    i,
                    medal["grade"],
                    medal["line"],
                    medal["map"],
                    medal["condition"],
                    medal["star"],
                )
            )

    return Medals(treasures, characters, actions, stages)


def medals(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting meow medals"""

    medal_stats = save_stats["medals"]
    remove = (
        user_input_handler.colored_input(
            "메달을 추가하거나 제거하시겠습니까?(&a&/&r&):"
        )
        == "r"
    )

    names = get_medal_names(helper.check_data_is_jp(save_stats))
    if names is None:
        return save_stats
    helper.colored_list(names)

    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "메달 ID 입력(&all&, 예: &1&-&50& 또는 공백으로 구분된 ID(예: &5 4 7&)을 얻기 위해 모두 입력할 수 있음):"
        ),
        len(names) + 1,
    )
    if remove:
        medal_stats = remove_medals(medal_stats, ids)
    else:
        medal_stats = set_medals(medal_stats, ids)
    save_stats["medals"] = medal_stats
    print(f"메달이 성공적으로 {'추가됨' if not remove else '제거됨'} ")
    return save_stats



"""Get game data from the BCData GitHub repository."""
import os
from typing import Optional
import requests

from . import helper

URL = "https://raw.githubusercontent.com/fieryhenry/BCData/master/"


def download_file(
    game_version: str,
    pack_name: str,
    file_name: str,
    get_data: bool = True,
    print_progress: bool = True,
) -> bytes:
    """
    Downloads the file.

    Args:
        game_version (str): The game version to download from.
        pack_name (str): The pack name to download from.
        file_name (str): The file name to download.
        get_data (bool, optional): Whether to return the data. Defaults to True.
        print_progress (bool, optional): Whether to print the progress. Defaults to True.

    Returns:
        bytes: The data of the file.
    """

    path = helper.get_file(os.path.join("game_data", game_version, pack_name))
    file_path = os.path.join(path, file_name)
    if os.path.exists(file_path):
        if get_data:
            return helper.read_file_bytes(file_path)
        return b""

    if print_progress:
        helper.colored_text(
            f"게임 데이터 파일 다운로드 중 &{file_name}& ~에서 &{pack_name}& 게임 버전으로 &{game_version}&",
            helper.GREEN,
            helper.WHITE,
        )
    url = URL + game_version + "/" + pack_name + "/" + file_name
    response = requests.get(url)

    helper.create_dirs(path)
    helper.write_file_bytes(file_path, response.content)
    return response.content


def get_latest_versions() -> Optional[list[str]]:
    """
    Gets the latest versions of the game data.

    Returns:
        Optional[list[str]]: The latest versions of the game data.
    """
    try:
        response = requests.get(URL + "latest.txt")
    except requests.exceptions.ConnectionError:
        return None
    versions = response.text.splitlines()
    return versions


def get_latest_version(is_jp: bool) -> Optional[str]:
    """
    Gets the latest version of the game data.

    Args:
        is_jp (bool): Whether to get the japanese version.

    Returns:
        str: The latest version of the game data.
    """
    versions = get_latest_versions()
    if versions is None:
        return None
    if is_jp:
        return versions[1]
    else:
        return versions[0]


def get_file_latest(pack_name: str, file_name: str, is_jp: bool) -> Optional[bytes]:
    """
    Gets the latest version of the file.

    Args:
        pack_name (str): The pack name to find.
        file_name (str): The file name to find.
        is_jp (bool): Whether to get the japanese version.

    Returns:
        Optional[bytes]: The data of the file.
    """
    version = get_latest_version(is_jp)
    if version is None:
        return None
    return download_file(version, pack_name, file_name)


def get_file_latest_path(path: str, is_jp: bool) -> Optional[bytes]:
    """
    Gets the latest version of the file.

    Args:
        path (str): The path to find.
        is_jp (bool): Whether to get the japanese version.

    Returns:
        Optional[bytes]: The data of the file.
    """
    version = get_latest_version(is_jp)
    if version is None:
        return None
    packname, filename = path.split("/")
    return download_file(version, packname, filename)


def get_path(pack_name: str, file_name: str, is_jp: bool) -> Optional[str]:
    """
    Gets the path of the file.

    Args:
        pack_name (str): The pack name to find.
        file_name (str): The file name to find.
        is_jp (bool): Whether to get the japanese version.

    Returns:
        Optional[str]: The path of the file.
    """
    version = get_latest_version(is_jp)
    if version is None:
        return None
    return os.path.join("game_data", version, pack_name, file_name)


def check_remove(new_version: str, is_jp: bool):
    """
    Checks if older game data is downloaded, and deletes if out of date.

    Args:
        new_version (str): The new version.
        is_jp (bool): Whether to get the japanese version.
    """
    all_versions = helper.get_dirs(helper.get_file("game_data"))
    for version in all_versions:
        if is_jp:
            if "jp" not in version:
                continue
            if version != new_version:
                helper.delete_dir(helper.get_file(os.path.join("game_data", version)))
        else:
            if "jp" in version:
                continue
            if version != new_version:
                helper.delete_dir(helper.get_file(os.path.join("game_data", version)))


def check_remove_handler():
    """
    Checks if older game data is downloaded, and deletes if out of date.
    """

    versions = get_latest_versions()
    if versions is None:
        return None
    check_remove(versions[0], is_jp=False)
    check_remove(versions[1], is_jp=True)



import json
from typing import Any, Optional
from BCSFE_Python import game_data_getter, csv_handler, helper, user_input_handler


class RawOrbInfo:
    def __init__(
        self,
        orb_id: int,
        grade_id: int,
        effect_id: int,
        value: list[int],
        attribute_id: int,
    ):
        """Initialize the RawOrbInfo class

        Args:
            orb_id (int): The id of the orb
            grade_id (int): The id of the grade
            effect_id (int): The id of the effect
            value (list[int]): The value of the effect? idk
            attribute_id (int): The id of the attribute
        """
        self.orb_id = orb_id
        self.grade_id = grade_id
        self.effect_id = effect_id
        self.value = value
        self.attribute_id = attribute_id


class OrbInfo:
    def __init__(
        self,
        raw_orb_info: RawOrbInfo,
        grade: str,
        attribute: str,
        effect: str,
    ):
        """Initialize the OrbInfo class

        Args:
            raw_orb_info (RawOrbInfo): The raw orb info
            grade (str): The grade of the orb (e.g. "S")
            attribute (str): The attribute of the orb (e.g. "Red")
            effect (str): The effect of the orb (e.g. "Attack Up %@: %@")
        """
        self.raw_orb_info = raw_orb_info
        self.grade = grade
        self.attribute = attribute
        self.effect = effect

    def __str__(self) -> str:
        """Get the string representation of the OrbInfo

        Returns:
            str: The string representation of the OrbInfo
        """
        effect_text = self.effect.replace("%@", "{}")
        text = effect_text.format(self.grade, self.attribute).strip()
        return text

    def to_colortext(self) -> str:
        """Get the string representation of the OrbInfo with color

        Returns:
            str: The string representation of the OrbInfo with color
        """
        effect_text = self.effect.replace("%@", "&{}&")
        text = effect_text.format(self.grade, self.attribute).strip()
        return text

    @staticmethod
    def create_unknown(orb_id: int) -> "OrbInfo":
        """Create an unknown OrbInfo

        Args:
            orb_id (int): The id of the orb

        Returns:
            OrbInfo: The unknown OrbInfo
        """
        return OrbInfo(
            RawOrbInfo(orb_id, 0, 0, [], 0),
            "Unknown",
            "",
            "%@:%@",
        )


class OrbInfoList:
    equipment_data_file_name = "DataLocal/equipmentlist.json"
    grade_list_file_name = "DataLocal/equipmentgrade.csv"
    attribute_list_file_name = "resLocal/attribute_explonation.tsv"
    effect_list_file_name = "resLocal/equipment_explonation.tsv"

    def __init__(self, orb_info_list: list[OrbInfo]):
        """Initialize the OrbInfoList class

        Args:
            orb_info_list (list[OrbInfo]): The list of OrbInfo
        """
        self.orb_info_list = orb_info_list

    @staticmethod
    def create(is_jp: bool) -> Optional["OrbInfoList"]:
        """Create an OrbInfoList

        Args:
            is_jp (bool): Whether the game is in Japanese

        Returns:
            Optional[OrbInfoList]: The OrbInfoList
        """
        json_data_file = game_data_getter.get_file_latest_path(
            OrbInfoList.equipment_data_file_name, is_jp
        )
        grade_list_file = game_data_getter.get_file_latest_path(
            OrbInfoList.grade_list_file_name, is_jp
        )
        attribute_list_file = game_data_getter.get_file_latest_path(
            OrbInfoList.attribute_list_file_name, is_jp
        )
        equipment_list_file = game_data_getter.get_file_latest_path(
            OrbInfoList.effect_list_file_name, is_jp
        )
        if (
            json_data_file is None
            or grade_list_file is None
            or attribute_list_file is None
            or equipment_list_file is None
        ):
            return None
        json_data = json_data_file.decode("utf-8")
        grade_list = grade_list_file.decode("utf-8")
        attribute_list = attribute_list_file.decode("utf-8")
        equipment_list = equipment_list_file.decode("utf-8")
        raw_orbs = OrbInfoList.parse_json_data(json_data)
        orbs = OrbInfoList.load_names(
            raw_orbs, grade_list, attribute_list, equipment_list
        )
        return OrbInfoList(orbs)

    @staticmethod
    def parse_json_data(json_data: str) -> list[RawOrbInfo]:
        """Parse the json data of the equipment

        Args:
            json_data (str): The json data

        Returns:
            list[RawOrbInfo]: The list of RawOrbInfo
        """
        data: dict[str, Any] = json.loads(json_data)
        orb_info_list: list[RawOrbInfo] = []
        for id, orb in enumerate(data["ID"]):
            grade_id = orb["gradeID"]
            content = orb["content"]
            value = orb["value"]
            attribute = orb["attribute"]
            orb_info_list.append(RawOrbInfo(id, grade_id, content, value, attribute))
        return orb_info_list

    @staticmethod
    def load_names(
        raw_orb_info: list[RawOrbInfo],
        grade_data: str,
        attribute_data: str,
        effect_data: str,
    ) -> list[OrbInfo]:
        """Load the names of the equipment

        Args:
            raw_orb_info (list[RawOrbInfo]): The list of RawOrbInfo
            grade_data (str): Raw data of the grade list
            attribute_data (str): Raw data of the attribute list
            effect_data (str): Raw data of the effect list

        Returns:
            list[OrbInfo]: The list of OrbInfo
        """
        grade_csv = csv_handler.parse_csv(grade_data)
        attribute_tsv = csv_handler.parse_csv(attribute_data, "\t")
        effect_csv = csv_handler.parse_csv(effect_data, "\t")
        orb_info_list: list[OrbInfo] = []
        for orb in raw_orb_info:
            grade = grade_csv[orb.grade_id][3]
            attribute = attribute_tsv[orb.attribute_id][0]
            effect = effect_csv[orb.effect_id][0]
            orb_info_list.append(OrbInfo(orb, grade, attribute, effect))
        return orb_info_list

    def get_orb_info(self, orb_id: int) -> Optional[OrbInfo]:
        """Get the OrbInfo from the id

        Args:
            orb_id (int): The id of the orb

        Returns:
            Optional[OrbInfo]: The OrbInfo
        """
        if orb_id >= len(self.orb_info_list):
            return None
        return self.orb_info_list[orb_id]

    def get_orb_from_components(
        self,
        grade: str,
        attribute: str,
        effect: str,
    ) -> Optional[OrbInfo]:
        """Get the OrbInfo from the components

        Args:
            grade (str): The grade of the orb
            attribute (str): The attribute of the orb
            effect (str): The effect of the orb

        Returns:
            Optional[OrbInfo]: The OrbInfo
        """
        for orb in self.orb_info_list:
            if (
                orb.grade == grade
                and orb.attribute == attribute
                and orb.effect == effect
            ):
                return orb
        return None

    def get_orbs_from_component_fuzzy(
        self,
        grade: str,
        attribute: str,
        effect: str,
    ) -> list[OrbInfo]:
        """Get the OrbInfo from the components matching the first word of the effect and lowercased

        Args:
            grade (str): The grade of the orb
            attribute (str): The attribute of the orb
            effect (str): The effect of the orb

        Returns:
            list[OrbInfo]: The list of OrbInfo
        """
        orbs: list[OrbInfo] = []
        for orb in self.orb_info_list:
            if (
                (orb.grade.lower() == grade.lower() or grade == "*")
                and (orb.attribute.lower() == attribute.lower() or attribute == "*")
                and (
                    orb.effect.lower().split(" ")[0] == effect.lower().split(" ")[0]
                    or effect == "*"
                )
            ):
                orbs.append(orb)
        return orbs

    def get_all_grades(self) -> list[str]:
        """Get all the grades

        Returns:
            list[str]: The list of grades
        """
        return list(set([orb.grade for orb in self.orb_info_list]))

    def get_all_attributes(self) -> list[str]:
        """Get all the attributes

        Returns:
            list[str]: The list of attributes
        """
        return list(set([orb.attribute for orb in self.orb_info_list]))

    def get_all_effects(self) -> list[str]:
        """Get all the effects

        Returns:
            list[str]: The list of effects
        """
        return list(set([orb.effect for orb in self.orb_info_list]))


class SaveOrb:
    """Represents a saved orb in the save file"""

    def __init__(self, orb: OrbInfo, count: int):
        """Initialize the SaveOrb class

        Args:
            orb (OrbInfo): The OrbInfo
            count (int): The amount of the orb
        """
        self.count = count
        self.orb = orb


class SaveOrbs:
    def __init__(
        self,
        orbs: dict[int, SaveOrb],
        orb_info_list: OrbInfoList,
    ):
        """Initialize the SaveOrbs class

        Args:
            orbs (dict[int, SaveOrb]): The orbs
            orb_info_list (OrbInfoList): The orb info list
        """
        self.orbs = orbs
        self.orb_info_list = orb_info_list

    @staticmethod
    def from_save_stats(save_stats: dict[str, Any]) -> Optional["SaveOrbs"]:
        """Create a SaveOrbs from the save stats

        Args:
            save_stats (dict[str, Any]): The save stats

        Returns:
            Optional[SaveOrbs]: The SaveOrbs
        """
        is_jp = helper.is_jp(save_stats)
        orb_info_list = OrbInfoList.create(is_jp)
        if orb_info_list is None:
            return None
        orbs: dict[int, SaveOrb] = {}
        for orb_id, amount in save_stats["talent_orbs"].items():
            try:
                orb_info = orb_info_list.orb_info_list[int(orb_id)]
            except IndexError:
                orb_info = OrbInfo.create_unknown(int(orb_id))
            orbs[int(orb_id)] = SaveOrb(orb_info, amount)

        return SaveOrbs(orbs, orb_info_list)

    def print(self):
        """Print the orbs as a formatted list"""
        self.sort_orbs()
        helper.colored_text(
            f"총 현재 오브: &{sum([orb.count for orb in self.orbs.values()])}&"
        )
        helper.colored_text(f"총 전류 유형: &{len(self.orbs)}&")
        print("현재 오브:")
        for orb in self.orbs.values():
            helper.colored_text(f"&{orb.count}& {orb.orb.to_colortext()}")

    def sort_orbs(self):
        """Sort the orbs by attribute, effect, grade and id in that order with attribute being the most important"""
        orbs = list(self.orbs.values())
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.orb_id)
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.grade_id)
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.effect_id)
        orbs.sort(key=lambda orb: orb.orb.raw_orb_info.attribute_id)

    def edit(self):
        """Edit the orbs"""
        self.print()
        all_grades = self.orb_info_list.get_all_grades()
        all_grades = [grade.lower() for grade in all_grades]
        all_grades.sort()
        all_attributes = self.orb_info_list.get_all_attributes()
        all_attributes = [attribute.lower() for attribute in all_attributes]
        all_attributes.sort()
        all_effects = self.orb_info_list.get_all_effects()
        all_effects = [effect.lower().split(" ")[0] for effect in all_effects]
        all_effects.sort()

        all_grades_str = "&,& ".join(all_grades)
        all_attributes_str = "&,& ".join(all_attributes)
        all_effects_str = "&,& ".join(all_effects)

        help_text = f"""Help:
사용 가능한 등급: &{all_grades_str}&
사용 가능한 속성: &{all_attributes_str}&
사용 가능한 효과: &{all_effects_str}&
&참고: 모든 속성에 대해 모든 등급과 효과를 사용할 수 있는 것은 아닙니다.&
예시 입력:
    &aku& - &all aku& 오브들 선택
    &red s& - &s& 등급의 모든 빨간색 &orb 선택
    &에일리언 d 공격& - &공격력&을 증가시키는 &d& 등급의 &에일리언 &orb를 선택합니다.
전환할 수 있으므로 다음과 같은 작업도 수행할 수 있습니다.
    &d 에일리언 공격&
    &빨간색&
    &공격 d 에일리언&
&all& 오브를 선택하려면 다음을 입력하십시오.
    &*&
&여러 선택&을 수행하려면 다음과 같이 &쉼표&로 구분하십시오.
    &s 블랙 터프&,&d 레드 매시브&,&플로팅&
"""
        helper.colored_text(help_text)

        orb_input_selection = (
            input("오브 선택:").lower().replace("angle", "angel").split(",")
        )
        orb_selection: list[OrbInfo] = []

        for orb_input in orb_input_selection:
            grade = None
            attribute = None
            effect = None
            orb_input = orb_input.strip()
            parts = orb_input.split(" ")
            parts = [part for part in parts if part != ""]
            if len(parts) == 0:
                continue
            if parts[0] == "*":
                orb_selection = self.orb_info_list.orb_info_list
                break
            for available_grade in all_grades:
                if available_grade in parts:
                    grade = available_grade
                    break
            for available_attribute in all_attributes:
                if available_attribute in parts:
                    attribute = available_attribute
                    break
            for available_effect in all_effects:
                if available_effect in parts:
                    effect = available_effect
                    break
            if grade is None:
                grade = "*"
            if attribute is None:
                attribute = "*"
            if effect is None:
                effect = "*"
            orbs = self.orb_info_list.get_orbs_from_component_fuzzy(
                grade, attribute, effect
            )
            orb_selection.extend(orbs)

        orb_selection = list(set(orb_selection))
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.orb_id)
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.grade_id)
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.effect_id)
        orb_selection.sort(key=lambda orb: orb.raw_orb_info.attribute_id)

        print("선택한 오브:")
        for orb in orb_selection:
            helper.colored_text((orb.to_colortext()))

        individual = (
            input("오브의 양을 개별적으로 수정하시겠습니까? 아니면 한꺼번에? (i/a)") == "i"
        )
        if individual:
            for orb in orb_selection:
                orb_id = orb.raw_orb_info.orb_id
                try:
                    orb_count = self.orbs[orb_id].count
                except KeyError:
                    orb_count = 0
                orb_count = user_input_handler.colored_input(
                    f"금액을 무엇으로 설정하시겠습니까? {orb.to_colortext()} to? (현재 &{orb_count}&) (&q& 종료 하가위한):"
                )
                if orb_count == "q":
                    break
                orb_count = helper.check_int_max(orb_count)
                if orb_count is None:
                    continue

                self.orbs[orb_id] = SaveOrb(orb, orb_count)

        else:
            orb_count = user_input_handler.get_int(
                "선택한 오브의 수량을 얼마로 설정하시겠습니까?:"
            )
            orb_count = helper.clamp_int(orb_count)
            for orb in orb_selection:
                orb_id = orb.raw_orb_info.orb_id
                self.orbs[orb_id] = SaveOrb(orb, orb_count)

        self.print()

    def save(self, save_stats: dict[str, Any]):
        """Save the orbs to the save_stats

        Args:
            save_stats (dict[str, Any]): The save_stats to save the orbs to
        """
        for orb_id, orb in self.orbs.items():
            save_stats["talent_orbs"][orb_id] = orb.count


def edit_talent_orbs(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Edit the talent orbs

    Args:
        save_stats (dict[str, Any]): The save_stats to edit the orbs in

    Returns:
        dict[str, Any]: The edited save_stats
    """
    save_orbs = SaveOrbs.from_save_stats(save_stats)
    if save_orbs is None:
        print("오브를 로드하지 못했습니다.")
        return save_stats
    save_orbs.edit()
    save_orbs.save(save_stats)
    return save_stats



"""Handler for server save management functions"""
from typing import Any

from ... import helper, serialise_save, server_handler, user_info


def upload_metadata(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Upload the metadata to the game server"""

    _, save_stats = server_handler.meta_data_upload_handler(
        save_stats, helper.get_save_path()
    )
    return save_stats


def set_managed_items(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Set the managed items for the save stats"""

    data = server_handler.check_gen_token(save_stats)
    token = data["token"]
    save_stats = data["save_stats"]
    if token is None:
        helper.colored_text("토큰 생성 오류")
        return save_stats
    server_handler.update_managed_items(save_stats["inquiry_code"], token, save_stats)
    return save_stats


def handle_upload_error(inquiry_code: str):
    """Show an error message"""
    info = user_info.UserInfo(inquiry_code)
    info.set_auth_token("")
    info.set_password("")
    helper.colored_text(
        "저장 데이터를 업로드하는 중 오류가 발생했습니다.\n다시 시도하십시오. 오류가 지속되면 #bug-reports에 신고해 주세요."
    )


def save_and_upload(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Serialise the save data, and upload it to the game server"""

    save_data = serialise_save.start_serialize(save_stats)
    save_data = helper.write_save_data(
        save_data, save_stats["version"], helper.get_save_path(), False
    )
    upload_data = server_handler.upload_handler(save_stats, helper.get_save_path())
    if upload_data is None:
        handle_upload_error(save_stats["inquiry_code"])
        return save_stats
    upload_data, save_stats = upload_data
    inquiry_code = save_stats["inquiry_code"]
    if upload_data is None:
        handle_upload_error(inquiry_code)
        return save_stats
    if "transferCode" not in upload_data:
        handle_upload_error(inquiry_code)
        return save_stats
    else:
        helper.colored_text(f"이어하기 코드 : &{upload_data['transferCode']}&")
        helper.colored_text(f"확인 코드 : &{upload_data['pin']}&")

    return save_stats



"""ManagedItem class for the BCSFE editor."""

from enum import Enum
from typing import Any
import uuid
from . import server_handler


class DetailType(Enum):
    """Enum for the different types of details."""

    GET = "get"
    USE = "use"


class ManagedItemType(Enum):
    """Enum for the different types of managed items."""

    CATFOOD = "catfood"
    RARE_TICKET = "rareTicket"
    PLATINUM_TICKET = "platinumTicket"
    LEGEND_TICKET = "legendTicket"


class ManagedItem:
    """Managed item for backupmetadata"""

    def __init__(
        self, amount: int, detail_type: DetailType, managed_item_type: ManagedItemType
    ):
        self.amount = amount
        self.detail_type = detail_type
        self.managed_item_type = managed_item_type
        self.detail_code = str(uuid.uuid4())
        self.detail_created_at = server_handler.get_current_time() - 109600

    def to_dict(self) -> dict[str, Any]:
        """Convert the managed item to a dictionary."""

        data = {
            "amount": self.amount,
            "detailCode": self.detail_code,
            "detailCreatedAt": self.detail_created_at,
            "detailType": self.detail_type.value,
            "managedItemType": self.managed_item_type.value,
        }
        return data




"""Handler for editing cata shrine xp and level"""
from typing import Any, Optional

from ... import game_data_getter, helper, item, user_input_handler


def get_boundaries(is_jp: bool) -> Optional[list[int]]:
    """
    Returns the xp requirements for each level

    Args:
        is_jp (bool): If the save file is japanese

    Returns:
        list[int]: The xp requirements for each level
    """
    file_data = game_data_getter.get_file_latest("resLocal", "jinja_level.csv", is_jp)
    if file_data is None:
        helper.error_text("신전 레벨 데이터를 가져오지 못했습니다.")
        return None
    boundaries = file_data.decode("utf-8").splitlines()
    xp_requirements: list[int] = []
    counter = 0
    for line in boundaries:
        requirement = int(line.split(helper.get_text_splitter(is_jp))[0])
        counter += requirement
        xp_requirements.append(counter)
    return xp_requirements


def get_level_from_xp(shrine_xp: int, is_jp: bool) -> Optional[dict[str, Any]]:
    """
    Returns the level, max level and max xp from the given xp

    Args:
        shrine_xp (int): The xp of the shrine
        is_jp (bool): If the save file is japanese

    Returns:
        dict[str, Any]: The level, max level, and max xp
    """
    xp_requirements = get_boundaries(is_jp)
    if xp_requirements is None:
        return None
    level = 1
    for requirement in xp_requirements:
        if shrine_xp >= requirement:
            level += 1
    if level > len(xp_requirements):
        level = len(xp_requirements)
    return {
        "level": level,
        "max_level": len(xp_requirements),
        "max_xp": xp_requirements[-2],
    }


def get_xp_from_level(level: int, is_jp: bool) -> Optional[int]:
    """
    Returns the xp required to reach the given level

    Returns:
        _type_: int
    """
    xp_requirements = get_boundaries(is_jp)
    if xp_requirements is None:
        return None
    if level <= 1:
        shrine_xp = 0
    else:
        shrine_xp = xp_requirements[level - 2]
    return shrine_xp


def edit_shrine_xp(save_stats: dict[str, Any]) -> dict[str, Any]:
    """
    Edit the shrine xp of the save file

    Args:
        save_stats (dict[str, Any]): The save file stats

    Returns:
        dict[str, Any]: The edited save file stats
    """

    shrine_xp = save_stats["cat_shrine"]["xp_offering"]

    data = get_level_from_xp(shrine_xp, helper.check_data_is_jp(save_stats))
    if data is None:
        return save_stats
    level = data["level"]

    helper.colored_text(f"신사 경험치: &{shrine_xp}&\n레벨: &{level}&")
    raw = (
        user_input_handler.colored_input(
            "신사 경험치를 편집하시겠습니까?(&1&) 또는 레벨(&2&)?:"
        )
        == "1"
    )

    if raw:
        cat_shrine_xp = item.IntItem(
            name="신사 경험치",
            value=item.Int(shrine_xp),
            max_value=None,
        )
        cat_shrine_xp.edit()
        shrine_xp = int(cat_shrine_xp.get_value())
    else:
        shrine_level = item.IntItem(
            name="신사 레벨",
            value=item.Int(level),
            max_value=data["max_level"],
        )
        shrine_level.edit()
        shrine_xp = get_xp_from_level(
            int(shrine_level.get_value()), helper.check_data_is_jp(save_stats)
        )
    if shrine_xp is None:
        return save_stats
    shrine_data = get_level_from_xp(shrine_xp, helper.check_data_is_jp(save_stats))
    if shrine_data is None:
        return save_stats
    shrine_level = shrine_data["level"]
    if shrine_level > data["max_level"]:
        shrine_level = data["max_level"]
    save_stats["shrine_dialogs"]["Value"] = shrine_level - 1  # Level up dialog
    save_stats["shrine_gone"] = 0
    save_stats["cat_shrine"]["stamp_1"] = 0
    save_stats["cat_shrine"]["stamp_2"] = 0

    save_stats["cat_shrine"]["xp_offering"] = shrine_xp
    return save_stats



"""Handler for upgrading the blue upgrades"""
from typing import Any

from ... import helper, user_input_handler
from . import upgrade_cats

TYPES = [
    "냥코 대포 공격력",
    "냥코 대포 사정거리",
    "냥코 대포 충전",
    "일 고양이의 의 호율",
    "일 고양이의 의 지갑",
    "성 체력",
    "연구력",
    "회계력",
    "공부력",
    "통솔력",
]


def upgrade_blue_ids(save_stats: dict[str, Any], ids: list[int]) -> dict[str, Any]:
    """Upgrade blue upgrades for a set of ids"""

    save_stats["blue_upgrades"] = upgrade_cats.upgrade_handler(
        data=save_stats["blue_upgrades"],
        ids=ids,
        item_name="업그레이드",
        save_stats=save_stats,
    )
    save_stats = upgrade_cats.set_user_popups(save_stats)
    print("특수 기술을 성공적으로 설정")
    return save_stats


def upgrade_blue(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing blue upgrades"""

    levels = save_stats["blue_upgrades"]
    levels_removed = {
        "Base": [levels["Base"][0]] + levels["Base"][2:],
        "Plus": [levels["Plus"][0]] + levels["Plus"][2:],
    }

    levels_removed_formated: list[str] = []
    for base, plus in zip(levels_removed["Base"], levels_removed["Plus"]):
        levels_removed_formated.append(f"{base + 1}+{plus}")

    print("무엇을 업그레이드하고 싶습니까?:\n⚠️1만이상숫자넣지마십시요")
    helper.colored_list(TYPES, extra_data=levels_removed_formated)

    total = len(TYPES) + 1
    ids = user_input_handler.colored_input(
        f"{total}. &한 번에&\n1부터 숫자를 입력하세요.{total} (공백으로 구분된 여러 값을 입력하여 한 번에 여러 값을 편집할 수 있습니다.):"
    ).split(" ")
    ids = user_input_handler.create_all_list_not_inc(ids, 11)
    ids = helper.parse_int_list(ids, -1)
    new_ids: list[int] = []
    for blue_id in ids:
        if blue_id > 0:
            blue_id += 1
        new_ids.append(blue_id)
    ids = new_ids
    save_stats = upgrade_blue_ids(save_stats, ids)
    return save_stats



"""Handler for editting outbreaks"""
from typing import Any, Optional

from ... import user_input_handler, helper
from . import main_story


def get_available_chapters(outbreaks: dict[int, Any]) -> list[str]:
    """Get available chapters"""

    available_chapters: list[str] = []
    for chapter_index in outbreaks:
        if chapter_index > 2:
            chapter_index -= 1
        if chapter_index > 6:
            continue
        available_chapters.append(main_story.CHAPTERS[chapter_index])
    return available_chapters


def set_outbreak(
    chapter_data: dict[int, int], val_to_set: int, total: Optional[int] = None
) -> dict[int, int]:
    """Set a chapter of an outbreak"""
    if total is None:
        total = len(chapter_data)

    for level_id in range(total):
        chapter_data[level_id] = val_to_set
    return chapter_data


def set_outbreaks(
    outbreaks: dict[int, Any],
    current_outbreaks: dict[int, Any],
    ids: list[int],
    clear: bool = True,
) -> tuple[dict[int, Any], dict[int, Any]]:
    """Set outbreaks"""
    for chapter_id in ids:
        outbreaks[chapter_id] = set_outbreak(
            outbreaks[chapter_id], 1 if clear else 0, 48
        )
        if chapter_id in current_outbreaks:
            if clear:
                current_outbreaks[chapter_id] = {}
    return outbreaks, current_outbreaks


def edit_outbreaks(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting outbreaks"""

    outbreaks = save_stats["outbreaks"]
    current_outbreaks = save_stats["current_outbreaks"]

    clear = (
        user_input_handler.colored_input(
            "좀비스테이지를 지우거나 지우지 않으시겠습니까? (&c&/&u&): "
        )
        == "c"
    )

    available_chapters = get_available_chapters(outbreaks)

    print("편집할 챕터:")
    ids = user_input_handler.select_not_inc(
        options=available_chapters,
        mode="에 대한 발병을 제거합니까?",
    )
    ids = helper.check_clamp(ids, len(available_chapters) + 1, 0, 0)
    ids = main_story.format_story_ids(ids)
    outbreaks, current_outbreaks = set_outbreaks(
        outbreaks, current_outbreaks, ids, clear
    )
    save_stats["outbreaks"] = outbreaks
    save_stats["current_outbreaks"] = current_outbreaks
    print("청소를 성공적으로 설정")
    return save_stats



"""Handler for miscalanous save management functions"""

from typing import Any

from ... import adb_handler, helper


def export(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Export the save stats to a json file"""

    helper.export_json(save_stats, helper.get_save_path() + ".json")

    return save_stats


def clear_data(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Clear data wrapper for the clear_data function"""

    confirm = input("데이터를 지우고 싶습니까? (y/n)?:").lower()
    if confirm == "y":
        adb_handler.adb_clear_save_data(save_stats["version"])
        print("데이터 삭제됨")

    return save_stats



from typing import Any

from ... import item, csv_handler, game_data_getter, helper


def get_base_mats_names(is_jp: bool) -> list[str]:
    """Get the base material names"""

    file_data = game_data_getter.get_file_latest("resLocal", "GatyaitemName.csv", is_jp)
    if file_data is None:
        helper.error_text("기본 재료 이름을 가져오지 못했습니다.")
        return []
    item_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        delimeter=helper.get_text_splitter(is_jp),
    )
    file_data = game_data_getter.get_file_latest("DataLocal", "Gatyaitembuy.csv", is_jp)
    if file_data is None:
        helper.error_text("재료 아이템 구매 데이터를 가져오지 못했습니다.")
        return []
    all_items = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
        )
    )[1:]
    base_mat_indexes: dict[int, str] = {}
    for item_id, item in enumerate(all_items):
        if item[6] == 7:
            index = int(item[7])
            base_mat_indexes[index] = item_names[item_id][0]

    base_mats_names: list[str] = []
    for index in sorted(base_mat_indexes):
        base_mats_names.append(base_mat_indexes[index])

    return base_mats_names


def edit_base_mats(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing base materials"""

    base_mats = item.IntItemGroup.from_lists(
        names=get_base_mats_names(helper.check_data_is_jp(save_stats)),
        values=save_stats["base_materials"],
        maxes=9999,
        group_name="오토토 재료",
    )
    base_mats.edit()
    save_stats["base_materials"] = base_mats.get_values()
    return save_stats



"""Handler to edit cat talents"""
from typing import Any, Optional

from ... import helper, item, csv_handler, game_data_getter, user_input_handler
from . import cat_id_selector


def get_talent_data(save_stats: dict[str, Any]) -> Optional[dict[Any, Any]]:
    """Get talent data for all cats"""

    file_data = game_data_getter.get_file_latest(
        "DataLocal", "SkillAcquisition.csv", helper.check_data_is_jp(save_stats)
    )
    if file_data is None:
        helper.error_text("본능 데이터를 가져오지 못했습니다.")
        return None
    talent_data_raw = helper.parse_int_list_list(
        csv_handler.parse_csv(
            file_data.decode("utf-8"),
        )
    )
    file_data = game_data_getter.get_file_latest(
        "resLocal", "SkillDescriptions.csv", helper.check_data_is_jp(save_stats)
    )
    if file_data is None:
        helper.error_text("본능 이름을 가져오지 못했습니다.")
        return None
    talent_names = csv_handler.parse_csv(
        file_data.decode("utf-8"),
        helper.get_text_splitter(helper.check_data_is_jp(save_stats)),
    )
    columns = helper.int_to_str_ls(talent_data_raw[0])
    new_talent_data: dict[Any, Any] = {}
    for j in range(1, len(talent_data_raw)):
        data = talent_data_raw[j]
        cat_id: int = int(data[0])
        new_talent_data[cat_id] = {}

        for data_i, column in zip(data, columns):
            new_talent_data = replace_name(
                cat_id=cat_id,
                column=column,
                data=data_i,
                talent_names=talent_names,
                new_data=new_talent_data,
            )
    return new_talent_data


def replace_name(
    cat_id: int,
    column: str,
    data: int,
    talent_names: list[list[str]],
    new_data: dict[Any, Any],
) -> dict[str, Any]:
    """Replace the text ids with the corresponding names"""

    new_data[cat_id][column] = data
    if (
        "textID" in column or "tFxtID_F" in column
    ):  # ponos made a typo, should be textID_F
        new_data[cat_id][column] = talent_names[data][1]
        stop_at = "<br>"
        if stop_at in new_data[cat_id][column]:
            index = new_data[cat_id][column].index(stop_at)
            new_data[cat_id][column] = new_data[cat_id][column][:index]
    return new_data


def find_order(
    cat_talents: list[dict[str, Any]], cat_talent_data: dict[str, Any]
) -> list[str]:
    """Find what talent slot each letter corresponds to"""

    letters = ["A", "B", "C", "D", "E", "F", "G", "H"]
    letter_order: list[str] = []

    for talent in cat_talents:
        talent_id = talent["id"]
        for letter in letters:
            key = f"abilityID_{letter}"
            if key not in cat_talent_data:
                continue
            ability_id = int(cat_talent_data[key])
            if ability_id == talent_id:
                letter_order.append(letter)
                break
    return letter_order


def get_cat_talents(
    cat_talents: list[dict[str, Any]], cat_talent_data: dict[str, Any]
) -> dict[Any, Any]:
    """Get the name and max value of each talent for a specific cat"""

    data: dict[Any, Any] = {}
    letter_order = find_order(cat_talents, cat_talent_data)
    for i, letter in enumerate(letter_order):
        cat_data = {}
        if letter == "F":
            text_id_str = "tFxtID_F"  # ponos made a typo, should be textID_F
        else:
            text_id_str = f"textID_{letter}"
        cat_data["name"] = cat_talent_data[text_id_str].strip("\n")
        cat_data["max"] = int(cat_talent_data[f"MAXLv_{letter}"])
        if cat_data["max"] == 0:
            cat_data["max"] = 1
        data[i] = cat_data
    return data


def get_talent_levels(
    talent_data: dict[int, Any], talents: dict[int, Any], cat_id: int
) -> list[int]:
    """Get the level of each talent for a specific cat"""

    cat_talent_data = talent_data[cat_id]
    cat_talents = talents[cat_id]
    cat_talent_data_formatted = get_cat_talents(cat_talents, cat_talent_data)
    cat_talents_levels: list[int] = []
    for talent_formatted in cat_talent_data_formatted.values():
        max_val = talent_formatted["max"]
        cat_talents_levels.append(max_val)
    return cat_talents_levels


def max_all_talents(save_stats: dict[str, Any]):
    """Max all talents for all cats"""
    max_all = (
        user_input_handler.colored_input(
            "Do you want to max talents or reset talents? (&m&/&r&):"
        )
        == "m"
    )
    if not max_all:
        return remove_all_talents(save_stats)
    talents = save_stats["talents"]

    ids = cat_id_selector.select_cats(save_stats)

    talent_data = get_talent_data(save_stats)
    if talent_data is None:
        return save_stats
    cat_talents_levels: list[int] = []
    for cat_id in ids:
        if cat_id not in talents or cat_id not in talent_data:
            continue
        cat_talents = talents[cat_id]
        cat_talents_levels = get_talent_levels(talent_data, talents, cat_id)
        for i, cat_talent_level in enumerate(cat_talents_levels):
            cat_talents[i]["level"] = cat_talent_level
        save_stats["talents"] = talents

    print("본능을 성공적으로 설정")
    return save_stats


def remove_all_talents(save_stats: dict[str, Any]) -> dict[str, Any]:
    """
    Remove all talents for all cats

    Args:
        save_stats (dict[str, Any]): The save stats

    Returns:
        dict[str, Any]: The save stats
    """
    talents = save_stats["talents"]

    ids = cat_id_selector.select_cats(save_stats)

    talent_data = get_talent_data(save_stats)
    if talent_data is None:
        return save_stats
    cat_talents_levels: list[int] = []
    for cat_id in ids:
        if cat_id not in talents or cat_id not in talent_data:
            continue
        cat_talents = talents[cat_id]
        cat_talents_levels = get_talent_levels(talent_data, talents, cat_id)
        for i in range(len(cat_talents_levels)):
            cat_talents[i]["level"] = 0
        save_stats["talents"] = talents

    print("본능을 성공적으로 제거했습니다.")
    return save_stats


def edit_talents_individual(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing talents"""

    talents = save_stats["talents"]
    ids = cat_id_selector.select_cats(save_stats)

    talent_data = get_talent_data(save_stats)
    if talent_data is None:
        return save_stats
    for cat_id in ids:
        cat_talents_levels: list[int] = []
        if cat_id not in talents or cat_id not in talent_data:
            # don't spam the user with messages if they selected alot of ids at once
            if len(ids) < 20:
                helper.colored_text(
                    f"Error cat &{cat_id}& does not have any talents",
                    helper.RED,
                    helper.WHITE,
                )
            continue
        cat_talent_data = talent_data[cat_id]
        cat_talents = talents[cat_id]
        cat_talent_data_formatted = get_cat_talents(cat_talents, cat_talent_data)
        names: list[str] = []
        maxes: list[int] = []
        for talent_index, cat_talent_formatted in cat_talent_data_formatted.items():
            names.append(cat_talent_formatted["name"])
            cat_talents_levels.append(cat_talents[talent_index]["level"])
            maxes.append(cat_talent_formatted["max"])
        helper.colored_text(f"Cat &{cat_id}& is selected:")
        cat_talents_levels_g = item.IntItemGroup.from_lists(
            names=names,
            values=cat_talents_levels,
            maxes=maxes,
            group_name="Talents",
        )
        cat_talents_levels_g.edit()
        cat_talents_levels = cat_talents_levels_g.get_values()
        for i, cat_talent_level in enumerate(cat_talents_levels):
            cat_talents[i]["level"] = cat_talent_level

        talents[cat_id] = cat_talents

        save_stats["talents"] = talents

    print("본능을 성공적으로 설정")
    return save_stats



"""Handler for editing gamatoto helpers"""
from typing import Any, Optional

from ... import item, game_data_getter, helper


def get_gamatoto_helpers(is_jp: bool) -> Optional[dict[str, Any]]:
    """Get the rarities of all gamatoto helpers"""

    if is_jp:
        country_code = "ja"
    else:
        country_code = "en"

    file_data = game_data_getter.get_file_latest(
        "resLocal", f"GamatotoExpedition_Members_name_{country_code}.csv", is_jp
    )
    if file_data is None:
        helper.error_text("가마토토 도우미 데이터를 가져오지 못했습니다.")
        return None
    data = file_data.decode("utf-8").splitlines()[1:]
    helpers: dict[str, Any] = {}
    for line in data:
        line_data = line.split(helper.get_text_splitter(is_jp))
        if len(line_data) < 5:
            break

        helper_id = line_data[0]
        rarity = int(line_data[1])
        type_str = line_data[4]
        helpers[helper_id] = {"Rarity_id": rarity, "Rarity_name": type_str}
    return helpers


def generate_helpers(user_input: list[int], helper_data: dict[str, Any]) -> list[int]:
    """Generate unique helpers from amounts of each"""

    final_helpers: list[int] = []
    values = list(helper_data.values())
    for i, usr_input in enumerate(user_input):
        for j, value in enumerate(values):
            if value["Rarity_id"] == i:
                final_helpers += list(range(j + 1, j + 1 + usr_input))
                break
    return final_helpers


def get_helper_rarities(helper_data: dict[str, Any]) -> list[str]:
    """Get the rarities of all gamatoto helpers"""

    rarities: list[str] = []
    for helpers in helper_data.values():
        if helpers["Rarity_name"] not in rarities:
            rarities.append(helpers["Rarity_name"])
    return rarities


def get_helpers(helpers: list[int], helper_data: dict[str, Any]) -> dict[str, Any]:
    """Get the amount of each type of helper"""

    current_helpers: dict[int, Any] = {}

    rarities = get_helper_rarities(helper_data)
    helper_count: dict[str, int] = {}
    for rarity in rarities:
        helper_count[rarity] = 0

    for helper_id in helpers:
        if helper_id == 0xFFFFFFFF:
            break
        current_helpers[helper_id] = helper_data[str(helper_id)]
        helper_count[current_helpers[helper_id]["Rarity_name"]] += 1
    return helper_count


def add_empty_helper_slots(helpers: list[int], final_helpers: list[int]):
    """Add empty helper slots to the end of the list"""

    empty_slots = len(helpers) - len(final_helpers)
    if empty_slots > 0:
        final_helpers += [0xFFFFFFFF] * empty_slots
    return final_helpers


def edit_helpers(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for gamatoto helpers"""

    helpers = save_stats["helpers"]
    helper_data = get_gamatoto_helpers(helper.check_data_is_jp(save_stats))
    if helper_data is None:
        return save_stats

    helper_count = get_helpers(helpers, helper_data)

    helpers_counts_input = item.IntItemGroup.from_lists(
        names=list(helper_count.keys()),
        values=list(helper_count.values()),
        group_name="가마토토 도우미",
        maxes=10,
    )
    helpers_counts_input.edit()
    final_helpers = generate_helpers(helpers_counts_input.get_values(), helper_data)
    helpers = add_empty_helper_slots(helpers, final_helpers)
    save_stats["helpers"] = helpers
    return save_stats



"""Handler for creating a new account"""

from typing import Any

from . import fix_elsewhere
from ... import helper, server_handler


def create_new_account(save_stats: dict[str, Any]):
    """Create a new account"""

    helper.colored_text("새 조회 코드 및 토큰 생성 중...", helper.GREEN)

    save_stats["inquiry_code"] = server_handler.get_inquiry_code()
    save_stats["token"] = "0" * 40
    save_stats = fix_elsewhere.fix_elsewhere(save_stats, force_mi=True)

    return save_stats


def create_new_account_no_input(save_stats: dict[str, Any]) -> dict[str, Any]:
    """
    Create a new account without asking for input

    Args:
        save_stats (dict[str, Any]): The save stats

    Returns:
        dict[str, Any]: The save stats
    """
    helper.colored_text("새 조회 코드 및 토큰 생성 중...", helper.GREEN)

    save_stats["inquiry_code"] = server_handler.get_inquiry_code()
    save_stats["token"] = "0" * 40
    save_stats = fix_elsewhere.fix_elsewhere(save_stats, force_mi=True, text=False)

    return save_stats



"""Setup for installing the package."""

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("src/modkr/files/version.txt", "r", encoding="utf-8") as fh:
    version = fh.read()

setuptools.setup(
    name="Koreditor",
    version=version,
    author="상어",
    description="냥코에디터 한글화",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sharkwodm/koreditor",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.9",
    install_requires=[
        "colored==1.4.4",
        "tk",
        "python-dateutil",
        "requests",
        "pyyaml",
    ],
    include_package_data=True,
    extras_require={
        "testing": [
            "pytest",
            "pytest-cov",
        ],
    },
    package_data={"modkr": ["py.typed"]},
    flake8={"max-line-length": 160},
)



"""Handler for editing catnip missions"""
from typing import Any, Optional

from ... import user_input_handler, game_data_getter, csv_handler, helper


def get_mission_conditions(is_jp: bool) -> Optional[dict[Any, Any]]:
    """Get the mission data and what you need to do to complete it"""

    file_data = game_data_getter.get_file_latest(
        "DataLocal", "Mission_Condition.csv", is_jp
    )
    if file_data is None:
        helper.error_text("미션 조건을 얻지 못했습니다.")
        return None
    mission_condition_data = file_data.decode("utf-8")
    mission_conditions_list = helper.parse_int_list_list(
        csv_handler.parse_csv(mission_condition_data)
    )
    mission_conditions: dict[Any, Any] = {}
    for line in mission_conditions_list[1:]:
        mission_id = line[0]
        mission_conditions[mission_id] = {
            "mission_type": line[1],
            "conditions_type": line[2],
            "progress_count": line[3],
            "conditions_value": line[4:],
        }
    return mission_conditions


def get_mission_names(is_jp: bool) -> Optional[dict[int, Any]]:
    """Get all mission names"""

    file_data = game_data_getter.get_file_latest("resLocal", "Mission_Name.csv", is_jp)
    if file_data is None:
        helper.error_text("임무 이름을 가져오지 못했습니다.")
        return None
    mission_name = file_data.decode("utf-8")
    mission_name_list = mission_name.split("\n")
    mission_names: dict[int, Any] = {}
    for mission_name in mission_name_list:
        line_data = mission_name.split(helper.get_text_splitter(is_jp))
        if helper.check_int(line_data[0]) is None:
            continue
        mission_id = int(line_data[0])
        name = line_data[1]
        name = name.replace("&", "\\&")
        mission_names[mission_id] = name
    return mission_names


def get_mission_names_from_ids(
    ids: list[int], mission_names: dict[int, Any]
) -> list[str]:
    """Get the mission names from the ids"""

    names: list[str] = []
    for mission_id in ids:
        if mission_id in mission_names:
            names.append(mission_names[mission_id])
    return names


def get_mission_ids(
    missions: dict[str, Any], conditions: dict[int, Any], names: dict[int, Any]
) -> tuple[list[int], list[str]]:
    """Get the mission ids and names from the conditions"""

    mission_ids_to_use: list[int] = []
    for mission_id in missions["states"]:
        if mission_id in conditions:
            mission_ids_to_use.append(mission_id)

    names_to_use = get_mission_names_from_ids(mission_ids_to_use, names)
    return mission_ids_to_use, names_to_use


def set_missions(
    missions: dict[str, Any],
    ids: list[int],
    conditions: dict[Any, Any],
    mission_ids_to_use: list[int],
    re_claim: bool,
) -> dict[str, Any]:
    """Set the missions"""

    for mission_id in ids:
        mission_id = helper.clamp(mission_id, 1, len(mission_ids_to_use))
        mission_id = mission_ids_to_use[mission_id]
        if re_claim:
            claim = True
        elif not re_claim and missions["states"][mission_id] != 4:
            claim = True
        else:
            claim = False
        if claim:
            missions["states"][mission_id] = 2
            missions["requirements"][mission_id] = conditions[mission_id][
                "progress_count"
            ]
    return missions


def edit_missions(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editting catnip missions"""

    missions = save_stats["missions"]

    names = get_mission_names(helper.check_data_is_jp(save_stats))
    conditions = get_mission_conditions(helper.check_data_is_jp(save_stats))

    if names is None or conditions is None:
        return save_stats

    mission_ids_to_use, names_to_use = get_mission_ids(missions, conditions, names)

    ids = user_input_handler.select_not_inc(
        options=names_to_use,
        mode="완벽한",
    )
    re_claim = (
        user_input_handler.colored_input(
            "이미 요청한 미션 &(1)&(보상을 다시 받을 수 있음)을 다시 완료하시겠습니까, 아니면 요청하지 않은 미션만 완료하시겠습니까&(2)&:"
        )
        == "1"
    )
    missions = set_missions(missions, ids, conditions, mission_ids_to_use, re_claim)
    save_stats["missions"] = missions
    print("성공적으로 완료된 임무")
    return save_stats



"""Handler for clearing behemoth culling stages"""
from typing import Any

from . import event_stages
from ... import user_input_handler


def edit_behemoth_culling(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for clearing behemoth culling stages"""

    stage_data = save_stats["behemoth_culling"]
    lengths = stage_data["Lengths"]

    ids = []
    ids = user_input_handler.get_range(
        user_input_handler.colored_input(
            "거대 컬링 ID 입력(예: &0& = &가프라의 숨겨진 숲&, &1& = &아시비니 사막&)(&all&을 입력하면 범위(예: 1-49, 또는 공백으로 구분된 ID(예: &5 4 7&)를 얻을 수 있음):"
        ),
        lengths["total"],
    )
    save_stats["behemoth_culling"] = event_stages.stage_handler(stage_data, ids, 0)
    return save_stats



"""Fix the elsewhere issue and unban an account"""
import json
import os
from typing import Any


from ... import helper, adb_handler, server_handler, user_info


def edit_cache(password: str, token: str, save_stats: dict[str, Any]) -> bool:
    """Edit the cache file in /data/data/jp.co.ponos.battlecats/files/cache/ to add the token and password"""

    data = {"password": password, "token": token}
    data_s = json.dumps(data)
    data_s = data_s.replace(" ", "")
    inquiry_code = save_stats["inquiry_code"]
    local_path = os.path.abspath(inquiry_code + ".json")

    helper.write_file_string(local_path, data_s)
    game_v = save_stats["version"]
    if game_v == "jp":
        game_v = ""
    try:
        success = adb_handler.run_adb_command(
            f'shell mv "{local_path}" "/data/data/jp.co.ponos.battlecats{game_v}/cache/{inquiry_code}.json"'
        )
    except adb_handler.ADBException:
        success = False
    os.remove(local_path)
    return success


def fix_elsewhere(
    save_stats: dict[str, Any], force_mi: bool = False, text: bool = True
) -> dict[str, Any]:
    """Handler for fixing the elsewhere issue and unban an account"""

    helper.colored_text("계정 비밀번호 얻기...", helper.GREEN)
    original_iq = save_stats["inquiry_code"]
    data = server_handler.check_gen_token(save_stats)
    token = data["token"]
    inquiry_code = data["inquiry_code"]
    if token is None:
        helper.colored_text("인증 토큰을 가져오지 못했습니다.", helper.RED)
        return save_stats
    if original_iq != inquiry_code or force_mi:
        info = user_info.UserInfo(inquiry_code)
        info.clear_managed_items()
        server_handler.update_managed_items(
            save_stats["inquiry_code"], token, save_stats
        )
    if text:
        helper.colored_text(
            "완료!\n재생을 누르면 금지 메시지가 표시될 수 있습니다. 그럴 경우 재생을 다시 누르면 사라집니다.\n계속하려면 엔터키를 누르십시오...(변경 사항을 저장해야 합니다.)",
            helper.DARK_YELLOW,
        )
        input()
    return save_stats



"""Handler for basic, items that work in a common way"""
from typing import Any
from ... import item, managed_item


def edit_cat_food(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing cat food"""

    cat_food = item.IntItem(
        name="통조림",
        value=item.Int(save_stats["cat_food"]["Value"]),
        max_value=45000,
        bannable=item.Bannable(
            managed_item.ManagedItemType.CATFOOD, save_stats["inquiry_code"]
        ),
    )
    cat_food.edit()
    save_stats["cat_food"]["Value"] = cat_food.get_value()
    return save_stats


def edit_xp(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing xp"""

    experience = item.IntItem(
        name="경험치",
        value=item.Int(save_stats["xp"]["Value"]),
        max_value=99999999,
    )
    experience.edit()
    save_stats["xp"]["Value"] = experience.get_value()
    return save_stats


def edit_normal_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing normal tickets"""

    normal_tickets = item.IntItem(
        name="노말 티켓",
        value=item.Int(save_stats["normal_tickets"]["Value"]),
        max_value=2999,
    )
    normal_tickets.edit()
    save_stats["normal_tickets"]["Value"] = normal_tickets.get_value()
    return save_stats


def edit_rare_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing rare tickets"""

    rare_tickets = item.IntItem(
        name="레어 티켓",
        value=item.Int(save_stats["rare_tickets"]["Value"]),
        max_value=299,
        bannable=item.Bannable(
            inquiry_code=save_stats["inquiry_code"],
            work_around='&Instead of editing rare tickets directly, use the "Normal Ticket Max Trade Progress" conversion feature instead! It is much more safe.',
            type=managed_item.ManagedItemType.RARE_TICKET,
        ),
    )
    rare_tickets.edit()
    save_stats["rare_tickets"]["Value"] = rare_tickets.get_value()
    return save_stats


def edit_platinum_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing platinum tickets"""

    platinum_tickets = item.IntItem(
        name="플레티넘 티켓",
        value=item.Int(save_stats["platinum_tickets"]["Value"]),
        max_value=9,
        bannable=item.Bannable(
            inquiry_code=save_stats["inquiry_code"],
            work_around="&플래티넘 티켓을 편집하는 대신 플래티넘 조각를 편집하세요! 그들은 훨씬 더 안전합니다. 10 플래티넘 조각 = 1 플래티넘 티켓",
            type=managed_item.ManagedItemType.PLATINUM_TICKET,
        ),
    )
    platinum_tickets.edit()
    save_stats["platinum_tickets"]["Value"] = platinum_tickets.get_value()
    return save_stats


def edit_platinum_shards(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing platinum shards"""

    ticket_amount = save_stats["platinum_tickets"]["Value"]
    max_value = 99 - (ticket_amount * 10)
    platinum_shards = item.IntItem(
        name="플래티넘 조각",
        value=item.Int(save_stats["platinum_shards"]["Value"]),
        max_value=max_value,
    )
    platinum_shards.edit()
    save_stats["platinum_shards"]["Value"] = platinum_shards.get_value()
    return save_stats


def edit_np(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing np"""

    nyanko_points = item.IntItem(
        name="NP",
        value=item.Int(save_stats["np"]["Value"]),
        max_value=9999,
    )
    nyanko_points.edit()
    save_stats["np"]["Value"] = nyanko_points.get_value()
    return save_stats


def edit_leadership(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing leadership"""

    leadership = item.IntItem(
        name="리더쉽",
        value=item.Int(save_stats["leadership"]["Value"]),
        max_value=9999,
    )
    leadership.edit()
    save_stats["leadership"]["Value"] = leadership.get_value()
    return save_stats


def edit_battle_items(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing battle items"""

    battle_items = item.IntItemGroup.from_lists(
        names=[
            "스피드 업",
            "트레저 레이더",
            "고양이 도령",
            "야옹컴[cpu]",
            "고양이 박사",
            "스냥이퍼",
        ],
        values=save_stats["battle_items"],
        maxes=9999,
        group_name="배틀 아이템",
    )
    battle_items.edit()
    save_stats["battle_items"] = battle_items.get_values()

    return save_stats


def edit_engineers(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing ototo engineers"""

    engineers = item.IntItem(
        name="오토토 조수",
        value=item.Int(save_stats["engineers"]["Value"]),
        max_value=5,
    )
    engineers.edit()
    save_stats["engineers"]["Value"] = engineers.get_value()
    return save_stats


def edit_catamins(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing catamins"""

    catamins = item.IntItemGroup.from_lists(
        names=[
            "드링크 A",
            "드링크 B",
            "드링크 C",
        ],
        values=save_stats["catamins"],
        maxes=9999,
        group_name="드링크",
    )
    catamins.edit()
    save_stats["catamins"] = catamins.get_values()
    return save_stats


def edit_inquiry_code(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the inquiry code"""

    print(
        "⚠️경고: 문의 코드 편집은 자신이 무엇을 하고 있는지 알고 있는 경우에만 수행해야 합니다! 올바르게 수행하지 않으면 게임 내에서 다른 오류가 발생하기 때문입니다!"
    )
    inquiry_code = item.StrItem(
        name="문의 코드",
        value=save_stats["inquiry_code"],
    )
    inquiry_code.edit()
    save_stats["inquiry_code"] = inquiry_code.get_value()
    return save_stats


def edit_rare_gacha_seed(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the rare gacha seed"""

    rare_gacha_seed = item.IntItem(
        name="레어 가챠 씨앗",
        value=item.Int(save_stats["rare_gacha_seed"]["Value"], signed=False),
        max_value=None,
    )
    rare_gacha_seed.edit()
    save_stats["rare_gacha_seed"]["Value"] = rare_gacha_seed.get_value()
    return save_stats


def edit_unlocked_slots(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the amount of unlocked slots"""

    unlocked_slots = item.IntItem(
        name="잠금 해제된 슬롯",
        value=item.Int(save_stats["unlocked_slots"]["Value"]),
        max_value=len(save_stats["slots"]),
    )
    unlocked_slots.edit()
    save_stats["unlocked_slots"]["Value"] = unlocked_slots.get_value()
    return save_stats


def edit_token(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the password-refresh-token"""

    print(
        "⚠️경고: 토큰 편집은 수행 중인 작업을 알고 있는 경우에만 수행해야 합니다! 올바르게 수행하지 않으면 게임 내에서 다른 오류가 발생하기 때문입니다!"
    )
    token = item.StrItem(
        name="Token",
        value=save_stats["token"],
    )
    token.edit()
    save_stats["token"] = token.get_value()
    return save_stats


def edit_restart_pack(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for giving the restart pack"""

    save_stats["restart_pack"]["Value"] = 1
    print("재시작 팩을 성공적으로 제공했습니다.")
    return save_stats


def edit_challenge_battle(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the score of the challenge battle"""

    challenge_battle = item.IntItem(
        name="Challenge Battle",
        value=item.Int(save_stats["challenge"]["Score"]["Value"]),
        max_value=None,
    )
    challenge_battle.edit()
    save_stats["challenge"]["Score"]["Value"] = challenge_battle.get_value()
    save_stats["challenge"]["Cleared"]["Value"] = 1
    return save_stats


def edit_legend_tickets(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing legend tickets"""

    legend_tickets = item.IntItem(
        name="레전드 티켓",
        value=item.Int(save_stats["legend_tickets"]["Value"]),
        max_value=4,
        bannable=item.Bannable(
            inquiry_code=save_stats["inquiry_code"],
            type=managed_item.ManagedItemType.LEGEND_TICKET,
        ),
    )
    legend_tickets.edit()
    save_stats["legend_tickets"]["Value"] = legend_tickets.get_value()
    return save_stats


def edit_dojo_score(save_stats: dict[str, Any]) -> dict[str, Any]:
    """Handler for editing the dojo score"""

    if not save_stats["dojo_data"]:
        save_stats["dojo_data"] = {0: {0: 0}}

    dojo_score = item.IntItem(
        name="도장 점수",
        value=item.Int(save_stats["dojo_data"][0][0]),
        max_value=None,
    )
    dojo_score.edit()
    save_stats["dojo_data"][0][0] = dojo_score.get_value()
    return save_stats



from typing import Any

from ... import helper, user_input_handler


def convert_to_jp(save_stats: dict[str, Any]) -> dict[str, Any]:
    save_stats["version"] = "jp"
    save_stats["dst"] = False

    helper.colored_text("변환된 데이터 저장 jp", helper.GREEN)
    return save_stats


def convert_to_non_jp(save_stats: dict[str, Any], cc: str) -> dict[str, Any]:
    save_stats["version"] = cc
    save_stats["dst"] = True

    helper.colored_text(f"변환된 데이터 저장 {cc}", helper.GREEN)
    return save_stats


def convert(save_stats: dict[str, Any], version: str) -> dict[str, Any]:
    if version == "jp":
        return convert_to_jp(save_stats)
    else:
        return convert_to_non_jp(save_stats, version)


def convert_save(save_stats: dict[str, Any]) -> dict[str, Any]:
    gvs = ["en", "jp", "kr", "tw"]

    helper.colored_text(
        "경고: 이로 인해 문제가 발생할 수 있으며 두 앱의 버전이 동일해야 합니다(예: 둘 다 12.3)!",
        helper.RED,
    )

    if save_stats["version"] in gvs:
        gvs.remove(save_stats["version"])

    gv_index = (
        user_input_handler.select_single(
            gvs, title="저장을 다음으로 변환할 버전을 선택하십시오."
        )
        - 1
    )
    gv = gvs[gv_index]

    return convert(save_stats, gv)
